# Copyright 2024 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import ssl
from typing import Union, List, Tuple, Any, Optional, Sequence, Mapping, Dict

import grpc
import httpx
from orjson import orjson

from .constants.constants import PredictorProtocol, INFERENCE_CONTENT_LENGTH_HEADER
from .errors import UnsupportedProtocol, InvalidInput
from .logging import trace_logger as logger
from .protocol.grpc.grpc_predict_v2_pb2 import (
    ServerReadyResponse,
    ServerLiveResponse,
    ModelReadyResponse,
    ServerReadyRequest,
    ServerLiveRequest,
    ModelReadyRequest,
)
from .protocol.grpc.grpc_predict_v2_pb2_grpc import GRPCInferenceServiceStub
from .protocol.infer_type import InferRequest, InferResponse
from .utils.utils import is_v2, is_v1


class _UseClientDefault:
    """
    For `timeout=...` parameter we need to be able to indicate the default "unset" state, in a way that is distinctly
    different to using `None`.

    The default "unset" state indicates that whatever default is set on the
    client should be used. This is different to setting `None`, which
    explicitly disables the parameter, possibly overriding a client default.

    For example we use `timeout=USE_CLIENT_DEFAULT` in the `infer()` signature.
    Omitting the `timeout` parameter will send a request using whatever default
    timeout has been configured on the client. Including `timeout=None` will
    ensure no timeout is used.

    Note that user code shouldn't need to use the `USE_CLIENT_DEFAULT` constant,
    but it is used internally when a parameter is not included.
    """


USE_CLIENT_DEFAULT = _UseClientDefault()


class InferenceGRPCClient:
    """
    Asynchronous GRPC inference client. This feature is currently in alpha and may be subject to change.
    Note: This client uses a default retry config. To override, explicitly provide the 'method_config' in channel
    options or to disable retry set the channel option ("grpc.enable_retries", 0).
    {
        "methodConfig": [
            {
                # Apply retry to all methods
                "name": [{}],
                "retryPolicy": {
                    "maxAttempts": 3,
                    "initialBackoff": "0.1s",
                    "maxBackoff": "1s",
                    "backoffMultiplier": 2,
                    "retryableStatusCodes": ["UNAVAILABLE"],
                },
            }
        ]
    }
    :param url: Inference server url as a string.
    :param verbose: (optional) A boolean to enable verbose logging. Defaults to False.
    :param use_ssl: (optional) A boolean value indicating whether to use an SSL-enabled channel (True) or not (False).
                    If creds provided the client will use SSL-enabled channel regardless of the specified value.
    :param root_certificates: (optional) Path to the PEM-encoded root certificates file as a string, or None to
                              retrieve them from a default location chosen by gRPC runtime. If creds provided this
                              will be ignored.
    :param private_key: (optional) Path to the PEM-encoded private key file as a string or None if no private key
                        should be used. If creds provided this will be ignored.
    :param certificate_chain: (optional) Path to the PEM-encoded certificate chain file as a string or None if no
                              certificate chain should be used. If creds provided this will be ignored.
    :param creds: (optional) A ChannelCredentials instance for secure channel communication.
    :param channel_args: (optional) An list of key-value pairs (channel_arguments in gRPC Core runtime) to configure
                         the channel.
    :param timeout (optional) The maximum end-to-end time, in seconds, the request is allowed to take. By default,
                   client timeout is 60 seconds. To disable timeout explicitly set it to 'None'.
    :param retries (optional) The number of retries if the request fails. This will be ignored if retry policy is provided in the 'channel_args'.
    """

    def __init__(
        self,
        url: str,
        verbose: bool = False,
        use_ssl: bool = False,
        root_certificates: str = None,
        private_key: str = None,
        certificate_chain: str = None,
        creds: grpc.ChannelCredentials = None,
        channel_args: List[Tuple[str, Any]] = None,
        timeout: Optional[float] = 60,
        retries: Optional[int] = 3,
    ):

        # requires appending the port to the predictor host for gRPC to work
        if ":" not in url:
            port = 443 if use_ssl else 80
            url = f"{url}:{port}"
        # Default retry config
        service_config_json = json.dumps(
            {
                "methodConfig": [
                    {
                        # Apply retry to all methods
                        "name": [{}],
                        "retryPolicy": {
                            "maxAttempts": retries,
                            "initialBackoff": "0.1s",
                            "maxBackoff": "1s",
                            "backoffMultiplier": 2,
                            "retryableStatusCodes": ["UNAVAILABLE"],
                        },
                    }
                ]
            }
        )
        # Explicitly check "is not None" here to support passing an empty
        # list to specify setting no channel arguments.
        if channel_args is not None:
            channel_opt = channel_args
            if ("grpc.enable_retries", 0) not in channel_opt:
                is_exist = False
                for key, _ in channel_opt:
                    if key == "grpc.service_config":
                        is_exist = True
                        break
                if ("grpc.enable_retries", 1) not in channel_opt:
                    channel_opt.append(("grpc.enable_retries", 1))
                if not is_exist and retries > 0:
                    channel_opt.append(("grpc.service_config", service_config_json))
        else:
            # To specify custom channel_opt, see the channel_args parameter.
            channel_opt = [
                ("grpc.max_send_message_length", -1),
                ("grpc.max_receive_message_length", -1),
            ]
            if retries > 0:
                channel_opt.append(("grpc.enable_retries", 1))
                channel_opt.append(("grpc.service_config", service_config_json))

        if creds:
            self._channel = grpc.aio.secure_channel(url, creds, options=channel_opt)
        elif use_ssl:
            rc_bytes = pk_bytes = cc_bytes = None
            if root_certificates is not None:
                with open(root_certificates, "rb") as rc_fs:
                    rc_bytes = rc_fs.read()
            if private_key is not None:
                with open(private_key, "rb") as pk_fs:
                    pk_bytes = pk_fs.read()
            if certificate_chain is not None:
                with open(certificate_chain, "rb") as cc_fs:
                    cc_bytes = cc_fs.read()
            creds = grpc.ssl_channel_credentials(
                root_certificates=rc_bytes,
                private_key=pk_bytes,
                certificate_chain=cc_bytes,
            )
            self._channel = grpc.aio.secure_channel(url, creds, options=channel_opt)
        else:
            self._channel = grpc.aio.insecure_channel(url, options=channel_opt)
        self._client_stub = GRPCInferenceServiceStub(self._channel)
        self._verbose = verbose
        self._timeout = timeout

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def close(self):
        """
        Close the client. Any future calls to server
        will result in an Error.
        """
        await self._channel.close()

    async def infer(
        self,
        infer_request: InferRequest,
        timeout: Union[Optional[float], _UseClientDefault] = USE_CLIENT_DEFAULT,
        headers: Union[grpc.aio.Metadata, Sequence[Tuple[str, str]], None] = None,
    ) -> InferResponse:
        """
        Run asynchronous inference using the supplied inputs.
        :param infer_request: Inference input data as InferRequest or ModelInferRequest object.
        :param timeout: (optional) The maximum end-to-end time, in seconds, the request is allowed to take.
                        The default value is 60 seconds. To disable timeout explicitly set it to 'None'.
                        This will override the client's timeout.
        :param headers: (optional) Additional headers to be transmitted with the request.
        :return: Inference output as ModelInferResponse.
        :raises RPCError for non-OK-status response.
        """
        metadata = headers if headers is not None else tuple()

        if isinstance(infer_request, InferRequest):
            infer_request = infer_request.to_grpc()
        else:
            raise InvalidInput("Invalid input format")
        if self._verbose:
            logger.info(
                "metadata: {}\n infer_request: {}".format(metadata, infer_request)
            )

        try:
            response = await self._client_stub.ModelInfer(
                request=infer_request,
                metadata=metadata,
                timeout=(
                    self._timeout if isinstance(timeout, _UseClientDefault) else timeout
                ),
            )
            response = InferResponse.from_grpc(response)
            if self._verbose:
                logger.info("infer response: %s", response)
            return response
        except grpc.RpcError as rpc_error:
            logger.error("Failed to infer: %s", rpc_error, exc_info=True)
            raise rpc_error

    async def is_server_ready(
        self,
        timeout: Union[Optional[float], _UseClientDefault] = USE_CLIENT_DEFAULT,
        headers: Union[grpc.aio.Metadata, Sequence[Tuple[str, str]], None] = None,
    ) -> bool:
        """
        Get readiness of the inference server.
        :param timeout: (optional) The maximum end-to-end time, in seconds, the request is allowed to take.
                        The default value is 60 seconds. To disable timeout explicitly set it to 'None'.
                        This will override the client's timeout.
        :param headers: (optional) Additional headers to be transmitted with the request.
        :return: True if server is ready, False if server is not ready.
        :raises RPCError for non-OK-status response.
        """
        try:
            response: ServerReadyResponse = await self._client_stub.ServerReady(
                ServerReadyRequest(),
                timeout=(
                    self._timeout if isinstance(timeout, _UseClientDefault) else timeout
                ),
                metadata=headers,
            )
            if self._verbose:
                logger.info("Server ready response: %s", response)
            return response.ready
        except grpc.RpcError as rpc_error:
            logger.error("Failed to get server readiness: %s", rpc_error, exc_info=True)
            raise rpc_error

    async def is_server_live(
        self,
        timeout: Union[Optional[float], _UseClientDefault] = USE_CLIENT_DEFAULT,
        headers: Union[grpc.aio.Metadata, Sequence[Tuple[str, str]], None] = None,
    ) -> bool:
        """
        Get liveness of the inference server.
        :param timeout: (optional) The maximum end-to-end time, in seconds, the request is allowed to take.
                        The default value is 60 seconds. To disable timeout explicitly set it to 'None'.
                        This will override the client's timeout.
        :param headers: (optional) Additional headers to be transmitted with the request.
        :return: True if server is live, False if server is not live.
        :raises RPCError for non-OK-status response.
        """
        try:
            response: ServerLiveResponse = await self._client_stub.ServerLive(
                ServerLiveRequest(),
                timeout=(
                    self._timeout if isinstance(timeout, _UseClientDefault) else timeout
                ),
                metadata=headers,
            )
            if self._verbose:
                logger.info("Server live response: %s", response)
            return response.live
        except grpc.RpcError as rpc_error:
            logger.error("Failed to get server liveness: %s", rpc_error, exc_info=True)
            raise rpc_error

    async def is_model_ready(
        self,
        model_name: str,
        timeout: Union[Optional[float], _UseClientDefault] = USE_CLIENT_DEFAULT,
        headers: Union[grpc.aio.Metadata, Sequence[Tuple[str, str]], None] = None,
    ) -> bool:
        """
        Get readiness of the specified model.
        :param model_name:  The name of the model to check for readiness.
        :param timeout: (optional) The maximum end-to-end time, in seconds, the request is allowed to take.
                        The default value is 60 seconds. To disable timeout explicitly set it to 'None'.
                        This will override the client's timeout.
        :param headers: (optional) Additional headers to be transmitted with the request.
        :return: True if model is ready, False if model is not ready.
        :raises RPCError for non-OK-status response or specified model not found.
        """
        try:
            response: ModelReadyResponse = await self._client_stub.ModelReady(
                ModelReadyRequest(name=model_name),
                timeout=(
                    self._timeout if isinstance(timeout, _UseClientDefault) else timeout
                ),
                metadata=headers,
            )
            if self._verbose:
                logger.info("Model %s ready response: %s", model_name, response)
            return response.ready
        except grpc.RpcError as rpc_error:
            logger.error(
                "Failed to get readiness of the model with name %s: %s",
                model_name,
                rpc_error,
                exc_info=True,
            )
            raise rpc_error


class RESTConfig:
    """
    Configuration for REST inference client.

    :param transport (optional) An asynchronous transport class to use for sending requests over the network.
    :param protocol (optional) Inference server protocol as string or PredictorProtocol object. Defaults to V1 protocol.
    :param retries (optional) An integer value indicating the number of retries in case of ConnectError or
                   ConnectTimeout. Defaults to 3.
    :param timeout (optional) The maximum end-to-end time, in seconds, the request is allowed to take. By default,
                   client timeout is 60 seconds. To disable timeout explicitly set it to None.
    :param http2 (optional) A boolean indicating if HTTP/2 support should be enabled. Defaults to False.
    :param cert (optional) An SSL certificate used by the requested host to authenticate the client.
                Either a path to an SSL certificate file, or two-tuple of (certificate file, key file), or
                a three-tuple of (certificate file, key file, password).
    :param verify (optional) SSL certificates (a.k.a CA bundle) used to verify the identity of requested hosts.
                  Either True (default CA bundle), a path to an SSL certificate file, a ssl.SSLContext, or False
                  (which will disable verification).
    :param auth (optional) An authentication class to use when sending inference requests. Refer httpx
    :param verbose (optional) A boolean to enable verbose logging. Defaults to False.
    """

    def __init__(
        self,
        transport: httpx.AsyncBaseTransport = None,
        protocol: Union[str, PredictorProtocol] = "v1",
        retries: int = 3,
        http2: bool = False,
        timeout: Union[float, None, tuple, httpx.Timeout] = 60,
        cert=None,
        verify: Union[str, bool, ssl.SSLContext] = True,
        auth=None,
        verbose: bool = False,
    ):
        self.transport = transport
        self.protocol = (
            protocol.value if isinstance(protocol, PredictorProtocol) else protocol
        )
        self.retries = retries
        self.http2 = http2
        self.timeout = timeout
        self.cert = cert
        self.verify = verify
        self.retries = retries
        self.auth = auth
        self.transport = transport
        self.verbose = verbose
        if self.transport is None:
            httpx.AsyncHTTPTransport(
                retries=self.retries,
                http2=self.http2,
                cert=self.cert,
                verify=self.verify,
            )


class InferenceRESTClient:
    """
    Asynchronous REST inference client. This feature is currently in alpha and may be subject to change.
    :param config (optional) A RESTConfig object which contains client configurations.
    """

    def __init__(self, config: RESTConfig = None):
        self._config = RESTConfig() if config is None else config
        self._client = httpx.AsyncClient(
            transport=self._config.transport,
            http2=self._config.http2,
            timeout=self._config.timeout,
            auth=self._config.auth,
            verify=self._config.verify,
        )

    def _construct_url(
        self, base_url: Union[str, httpx.URL], relative_url: str
    ) -> httpx.URL:
        """
        Merge a relative url argument together with any 'base_url' to create the URL used for the outgoing request.
        :param base_url: The base url as str or httpx.URL object to use when constructing request url.
        :param relative_url: The relative url to use for merging with base url as string.
        :return: a httpx.URL object
        :raises InvalidURL if the base url is not valid.
        """
        if isinstance(base_url, str):
            base_url = httpx.URL(base_url)
        if base_url.scheme not in ("http", "https"):
            raise httpx.InvalidURL(
                "Base url should have 'http://' or 'https://' protocol"
            )
        if base_url.is_relative_url:
            raise httpx.InvalidURL("Base url should not be a relative url")
        if not base_url.raw_path.endswith(b"/") and not relative_url.startswith("/"):
            relative_url = "/" + relative_url
        return base_url.join(base_url.path + relative_url)

    def _construct_http_status_error(
        self, response: httpx.Response
    ) -> httpx.HTTPStatusError:
        message = (
            "{error_message}, '{0.status_code} {0.reason_phrase}' for url '{0.url}'"
        )
        error_message = ""
        if (
            "content-type" in response.headers
            and response.headers["content-type"] == "application/json"
        ):
            error_message = response.json()
            if "error" in error_message:
                error_message = error_message["error"]
        message = message.format(response, error_message=error_message)
        return httpx.HTTPStatusError(
            message, request=response.request, response=response
        )

    async def infer(
        self,
        base_url: Union[httpx.URL, str],
        data: Union[InferRequest, dict],
        model_name: Optional[str] = None,
        headers: Optional[Mapping[str, str]] = None,
        response_headers: Dict[str, str] = None,
        is_graph_endpoint: bool = False,
        timeout: Union[float, None, tuple, httpx.Timeout] = httpx.USE_CLIENT_DEFAULT,
    ) -> Union[InferResponse, Dict]:
        """
        Run asynchronous inference using the supplied data.
        :param base_url: Base url of the inference server. E.g. https://example.com:443, https://example.com:443/serving
        :param data: Input data as InferRequest object.
        :param model_name: (optional) Name of the model as string. If is_graph_endpoint is true this can be omitted.
               If is_graph_endpoint is False and model_name is None, this will raise ValueError.
        :param headers: (optional) HTTP headers to include when sending request.
        :param is_graph_endpoint: (optional) If set to True the base_url will be considered as an inference graph
                                  endpoint and will be used as it is for making the request regardless of the
                                  protocol specified in the RESTConfig. The result will be returned as dict object.
                                  Defaults to False.
        :param timeout: (optional) The maximum end-to-end time, in seconds, the request is allowed to take. This will
                        override the timeout in the RESTConfig. The default value is 60 seconds.
                        To disable timeout explicitly set it to 'None'.
        :return: Inference result as InferResponse object or python dict.
        :raises HTTPStatusError for response codes other than 2xx.
        :raises UnsupportedProtocol if the specified protocol version is not supported.
        """
        if is_graph_endpoint:
            url = base_url
        elif model_name is None:
            raise ValueError("model_name should not be 'None'")
        elif is_v1(self._config.protocol):
            url = self._construct_url(
                base_url, f"{self._config.protocol}/models/{model_name}:predict"
            )
        elif is_v2(self._config.protocol):
            url = self._construct_url(
                base_url, f"{self._config.protocol}/models/{model_name}/infer"
            )
        else:
            raise UnsupportedProtocol(self._config.protocol)
        if self._config.verbose:
            logger.info("url: %s", url)
            logger.info("request data: %s", data)
        if isinstance(data, InferRequest):
            data, json_length = data.to_rest()
            if json_length:
                headers = headers or {}
                headers[INFERENCE_CONTENT_LENGTH_HEADER] = str(json_length)
                headers["content-type"] = "application/octet-stream"
        if isinstance(data, dict):
            data = orjson.dumps(data)
        response = await self._client.post(
            url, content=data, headers=headers, timeout=timeout
        )
        if self._config.verbose:
            logger.info(
                "response code: %s, content: %s", response.status_code, response.text
            )
        if not response.is_success:
            raise self._construct_http_status_error(response)
        if response_headers is not None:
            response_headers.update(response.headers)
        # If inference graph result, return it as dict
        if is_graph_endpoint:
            output = orjson.loads(response.content)
        elif is_v2(self._config.protocol):
            json_length = response.headers.get(
                INFERENCE_CONTENT_LENGTH_HEADER, len(response.content)
            )
            output = InferResponse.from_bytes(response.content, int(json_length))
        # Should be v1 protocol result, return it as dict
        else:
            output = orjson.loads(response.content)
        return output

    async def explain(
        self,
        base_url: Union[httpx.URL, str],
        model_name: str,
        data: Dict,
        headers: Optional[Mapping[str, str]] = None,
        timeout: Union[float, None, tuple, httpx.Timeout] = httpx.USE_CLIENT_DEFAULT,
    ) -> Dict:
        """
        Run asynchronous explanation using the supplied data.
        :param base_url: Base url of the inference server. E.g. https://example.com:443, https://example.com:443/serving
        :param model_name: Name of the model as string.
        :param data: Input data as python dict.
        :param headers: (optional) HTTP headers to include when sending request.
        :param timeout: (optional) The maximum end-to-end time, in seconds, the request is allowed to take. This will
                        override the timeout in the RESTConfig. The default value is 60 seconds.
                        To disable timeout explicitly set it to 'None'.
        :return: Explain result as python dict.
        :raises HTTPStatusError for response codes other than 2xx.
        :raises UnsupportedProtocol if the specified protocol version is not supported.
        """
        if is_v1(self._config.protocol):
            url = self._construct_url(
                base_url, f"{self._config.protocol}/models/{model_name}:explain"
            )
        else:
            raise UnsupportedProtocol(self._config.protocol)
        if self._config.verbose:
            logger.info("url: %s", url)
            logger.info("request data: %s", data)
        data = orjson.dumps(data)
        response = await self._client.post(
            url, content=data, headers=headers, timeout=timeout
        )
        if self._config.verbose:
            logger.info(
                "response code: %s, content: %s", response.status_code, response.text
            )
        if not response.is_success:
            raise self._construct_http_status_error(response)
        return orjson.loads(response.content)

    async def is_server_ready(
        self,
        base_url: Union[httpx.URL, str],
        headers: Optional[Mapping[str, str]] = None,
        timeout: Union[float, None, tuple, httpx.Timeout] = httpx.USE_CLIENT_DEFAULT,
    ) -> bool:
        """
        Get readiness of the inference server.
        :param base_url: Base url of the inference server. E.g. https://example.com:443, https://example.com:443/serving
        :param headers: (optional) HTTP headers to include when sending request.
        :param timeout: (optional) The maximum end-to-end time, in seconds, the request is allowed to take. This will
                        override the timeout in the RESTConfig. The default value is 60 seconds.
                        To disable timeout explicitly set it to 'None'.
        :return: True if server is ready, False if server is not ready.
        :raises HTTPStatusError for response codes other than 2xx.
        :raises UnsupportedProtocol if the specified protocol version is not supported.
        """
        if is_v2(self._config.protocol):
            url = self._construct_url(base_url, f"{self._config.protocol}/health/ready")
        else:
            raise UnsupportedProtocol(protocol_version=self._config.protocol)
        if self._config.verbose:
            logger.info("url: %s, protocol_version: %s", url, self._config.protocol)
        response = await self._client.get(url, headers=headers, timeout=timeout)
        if self._config.verbose:
            logger.info(
                "response code: %s, content: %s", response.status_code, response.text
            )
        if not response.is_success:
            raise self._construct_http_status_error(response)
        return response.json().get("ready")

    async def is_server_live(
        self,
        base_url: Union[str, httpx.URL],
        headers: Optional[Mapping[str, str]] = None,
        timeout: Union[float, None, tuple, httpx.Timeout] = httpx.USE_CLIENT_DEFAULT,
    ) -> bool:
        """
        Get liveness of the inference server.
        :param base_url: Base url of the inference server. E.g. https://example.com:443, https://example.com:443/serving
        :param headers: (optional) HTTP headers to include when sending request.
        :param timeout: (optional) The maximum end-to-end time, in seconds, the request is allowed to take. This will
                        override the timeout in the RESTConfig. The default value is 60 seconds.
                        To disable timeout explicitly set it to 'None'.
        :return: True if server is ready, False if server is not ready.
        :raises HTTPStatusError for response codes other than 2xx.
        :raises UnsupportedProtocol if the specified protocol version is not supported.
        """
        if is_v1(self._config.protocol):
            url = self._construct_url(base_url, "")
        elif is_v2(self._config.protocol):
            url = self._construct_url(base_url, f"{self._config.protocol}/health/live")
        else:
            raise UnsupportedProtocol(protocol_version=self._config.protocol)
        if self._config.verbose:
            logger.info("url: %s, protocol_version: %s", url, self._config.protocol)
        response = await self._client.get(url, headers=headers, timeout=timeout)
        if self._config.verbose:
            logger.info(
                "response code: %s, content: %s", response.status_code, response.text
            )
        if not response.is_success:
            raise self._construct_http_status_error(response)
        if is_v1(self._config.protocol):
            is_live = response.json().get("status").lower() == "alive"
        elif is_v2(self._config.protocol):
            is_live = response.json().get("live")
        else:
            raise UnsupportedProtocol(protocol_version=self._config.protocol)
        return is_live

    async def is_model_ready(
        self,
        base_url: Union[httpx.URL, str],
        model_name: str,
        headers: Optional[Mapping[str, str]] = None,
        timeout: Union[float, None, tuple, httpx.Timeout] = httpx.USE_CLIENT_DEFAULT,
    ) -> bool:
        """
        Get readiness of the specified model.
        :param base_url: Base url of the inference server. E.g. https://example.com:443, https://example.com:443/serving
        :param model_name: Name of the model as string.
        :param headers: (optional) HTTP headers to include when sending request.
        :param timeout: (optional) The maximum end-to-end time, in seconds, the request is allowed to take. This will
                        override the timeout in the RESTConfig. The default value is 60 seconds.
                        To disable timeout explicitly set it to 'None'.
        :return: True if server is ready, False if server is not ready.
        :raises HTTPStatusError for response codes other than 2xx.
        :raises UnsupportedProtocol if the specified protocol version is not supported.
        """
        if is_v1(self._config.protocol):
            url = self._construct_url(
                base_url, f"{self._config.protocol}/models/{model_name}"
            )
        elif is_v2(self._config.protocol):
            url = self._construct_url(
                base_url, f"{self._config.protocol}/models/{model_name}/ready"
            )
        else:
            raise UnsupportedProtocol(protocol_version=self._config.protocol)
        if self._config.verbose:
            logger.info("url: %s, protocol_version: %s", url, self._config.protocol)
        response = await self._client.get(url, headers=headers, timeout=timeout)
        if self._config.verbose:
            logger.info(
                "response code: %s, content: %s", response.status_code, response.text
            )

        # Model Server responds with 503 service unavailable error if model is not ready
        if response.status_code == httpx.codes.SERVICE_UNAVAILABLE:
            return False
        # Raise for other status codes
        if not response.is_success:
            raise self._construct_http_status_error(response)
        return response.json().get("ready")

    async def close(self):
        """
        Close the client, transport and proxies.
        """
        await self._client.aclose()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Dict, Optional, Union

import uvicorn
from fastapi import FastAPI, Request, Response
from fastapi.routing import APIRouter
from prometheus_client import REGISTRY, exposition
from timing_asgi import TimingClient, TimingMiddleware
from timing_asgi.integrations import StarletteScopeToName

from kserve.errors import (
    InferenceError,
    InvalidInput,
    ModelNotFound,
    ModelNotReady,
    ServerNotLive,
    ServerNotReady,
    UnsupportedProtocol,
    generic_exception_handler,
    inference_error_handler,
    invalid_input_handler,
    model_not_found_handler,
    model_not_ready_handler,
    not_implemented_error_handler,
    server_not_live_handler,
    server_not_ready_handler,
    unsupported_protocol_error_handler,
)
from kserve.logging import trace_logger, logger
from kserve.protocol.dataplane import DataPlane

from .openai.config import maybe_register_openai_endpoints
from .v1_endpoints import register_v1_endpoints
from .v2_endpoints import register_v2_endpoints
from ..model_repository_extension import ModelRepositoryExtension


async def metrics_handler(request: Request) -> Response:
    encoder, content_type = exposition.choose_encoder(request.headers.get("accept"))
    return Response(content=encoder(REGISTRY), headers={"content-type": content_type})


class PrintTimings(TimingClient):
    def timing(self, metric_name, timing, tags):
        trace_logger.info(f"{metric_name}: {timing} {tags}")


class RESTServer:
    def __init__(
        self,
        app: FastAPI,
        data_plane: DataPlane,
        model_repository_extension: ModelRepositoryExtension,
        http_port: int,
        log_config: Optional[Union[str, Dict]] = None,
        access_log_format: Optional[str] = None,
        workers: int = 1,
    ):
        self.app = app
        self.dataplane = data_plane
        self.model_repository_extension = model_repository_extension
        self.access_log_format = access_log_format
        self._server = uvicorn.Server(
            config=uvicorn.Config(
                app="kserve.model_server:app",
                host="0.0.0.0",
                log_config=log_config,
                port=http_port,
                workers=workers,
            )
        )

    def _register_endpoints(self):
        root_router = APIRouter()
        root_router.add_api_route(r"/", self.dataplane.live)
        root_router.add_api_route(r"/metrics", metrics_handler, methods=["GET"])
        self.app.include_router(root_router)
        register_v1_endpoints(self.app, self.dataplane, self.model_repository_extension)
        register_v2_endpoints(self.app, self.dataplane, self.model_repository_extension)
        # Register OpenAI endpoints if any of the models in the registry implement the OpenAI interface
        # This adds /openai/v1/completions and /openai/v1/chat/completions routes to the
        # REST server.
        maybe_register_openai_endpoints(self.app, self.dataplane.model_registry)

    def _add_exception_handlers(self):
        self.app.add_exception_handler(InvalidInput, invalid_input_handler)
        self.app.add_exception_handler(InferenceError, inference_error_handler)
        self.app.add_exception_handler(ModelNotFound, model_not_found_handler)
        self.app.add_exception_handler(ModelNotReady, model_not_ready_handler)
        self.app.add_exception_handler(
            NotImplementedError, not_implemented_error_handler
        )
        self.app.add_exception_handler(
            UnsupportedProtocol, unsupported_protocol_error_handler
        )
        self.app.add_exception_handler(ServerNotLive, server_not_live_handler)
        self.app.add_exception_handler(ServerNotReady, server_not_ready_handler)
        self.app.add_exception_handler(Exception, generic_exception_handler)

    def _add_middlewares(self):
        self.app.add_middleware(
            TimingMiddleware,
            client=PrintTimings(),
            metric_namer=StarletteScopeToName(
                prefix="kserve.io", starlette_app=self.app
            ),
        )

        # More context in https://github.com/encode/uvicorn/pull/947
        # At the time of writing the ASGI specs are not clear when it comes
        # to change the access log format, and hence the Uvicorn upstream devs
        # chose to create a custom middleware for this.
        # The allowed log format is specified in https://github.com/Kludex/asgi-logger#usage
        if self.access_log_format:
            from asgi_logger import AccessLoggerMiddleware

            # As indicated by the asgi-logger docs, we need to clear/unset
            # any setting for uvicorn.access to avoid log duplicates.
            logging.getLogger("uvicorn.access").handlers = []
            self.app.add_middleware(
                AccessLoggerMiddleware, format=self.access_log_format
            )
            # The asgi-logger settings don't set propagate to False,
            # so we get duplicates if we don't set it explicitly.
            logging.getLogger("access").propagate = False

    def create_application(self):
        self._add_middlewares()
        self._register_endpoints()
        self._add_exception_handlers()

    async def start(self):
        self.create_application()
        logger.info(f"Starting uvicorn with {self._server.config.workers} workers")
        await self._server.serve()



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

# flake8: noqa
"""
    KServe

    Python SDK for KServe  # noqa: E501

    The version of the OpenAPI document: v0.1
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

# import models into model package
from kserve.models.v1alpha1_built_in_adapter import V1alpha1BuiltInAdapter
from kserve.models.v1alpha1_cluster_serving_runtime import V1alpha1ClusterServingRuntime
from kserve.models.v1alpha1_cluster_serving_runtime_list import V1alpha1ClusterServingRuntimeList
from kserve.models.v1alpha1_cluster_storage_container import V1alpha1ClusterStorageContainer
from kserve.models.v1alpha1_cluster_storage_container_list import V1alpha1ClusterStorageContainerList
from kserve.models.v1alpha1_inference_graph import V1alpha1InferenceGraph
from kserve.models.v1alpha1_inference_graph_list import V1alpha1InferenceGraphList
from kserve.models.v1alpha1_inference_graph_spec import V1alpha1InferenceGraphSpec
from kserve.models.v1alpha1_inference_graph_status import V1alpha1InferenceGraphStatus
from kserve.models.v1alpha1_inference_router import V1alpha1InferenceRouter
from kserve.models.v1alpha1_inference_step import V1alpha1InferenceStep
from kserve.models.v1alpha1_inference_target import V1alpha1InferenceTarget
from kserve.models.v1alpha1_local_model_cache import V1alpha1LocalModelCache
from kserve.models.v1alpha1_local_model_cache_list import V1alpha1LocalModelCacheList
from kserve.models.v1alpha1_local_model_cache_spec import V1alpha1LocalModelCacheSpec
from kserve.models.v1alpha1_local_model_node import V1alpha1LocalModelNode
from kserve.models.v1alpha1_local_model_node_group import V1alpha1LocalModelNodeGroup
from kserve.models.v1alpha1_local_model_node_group_list import V1alpha1LocalModelNodeGroupList
from kserve.models.v1alpha1_local_model_node_group_spec import V1alpha1LocalModelNodeGroupSpec
from kserve.models.v1alpha1_local_model_node_list import V1alpha1LocalModelNodeList
from kserve.models.v1alpha1_local_model_node_spec import V1alpha1LocalModelNodeSpec
from kserve.models.v1alpha1_model_spec import V1alpha1ModelSpec
from kserve.models.v1alpha1_serving_runtime import V1alpha1ServingRuntime
from kserve.models.v1alpha1_serving_runtime_list import V1alpha1ServingRuntimeList
from kserve.models.v1alpha1_serving_runtime_pod_spec import V1alpha1ServingRuntimePodSpec
from kserve.models.v1alpha1_serving_runtime_spec import V1alpha1ServingRuntimeSpec
from kserve.models.v1alpha1_storage_container_spec import V1alpha1StorageContainerSpec
from kserve.models.v1alpha1_storage_helper import V1alpha1StorageHelper
from kserve.models.v1alpha1_supported_model_format import V1alpha1SupportedModelFormat
from kserve.models.v1alpha1_supported_uri_format import V1alpha1SupportedUriFormat
from kserve.models.v1alpha1_trained_model import V1alpha1TrainedModel
from kserve.models.v1alpha1_trained_model_list import V1alpha1TrainedModelList
from kserve.models.v1alpha1_trained_model_spec import V1alpha1TrainedModelSpec
from kserve.models.v1beta1_art_explainer_spec import V1beta1ARTExplainerSpec
from kserve.models.v1beta1_batcher import V1beta1Batcher
from kserve.models.v1beta1_component_extension_spec import V1beta1ComponentExtensionSpec
from kserve.models.v1beta1_component_status_spec import V1beta1ComponentStatusSpec
from kserve.models.v1beta1_custom_explainer import V1beta1CustomExplainer
from kserve.models.v1beta1_custom_predictor import V1beta1CustomPredictor
from kserve.models.v1beta1_custom_transformer import V1beta1CustomTransformer
from kserve.models.v1beta1_deploy_config import V1beta1DeployConfig
from kserve.models.v1beta1_explainer_config import V1beta1ExplainerConfig
from kserve.models.v1beta1_explainer_extension_spec import V1beta1ExplainerExtensionSpec
from kserve.models.v1beta1_explainer_spec import V1beta1ExplainerSpec
from kserve.models.v1beta1_explainers_config import V1beta1ExplainersConfig
from kserve.models.v1beta1_failure_info import V1beta1FailureInfo
from kserve.models.v1beta1_hugging_face_runtime_spec import V1beta1HuggingFaceRuntimeSpec
from kserve.models.v1beta1_inference_service import V1beta1InferenceService
from kserve.models.v1beta1_inference_service_list import V1beta1InferenceServiceList
from kserve.models.v1beta1_inference_service_spec import V1beta1InferenceServiceSpec
from kserve.models.v1beta1_inference_service_status import V1beta1InferenceServiceStatus
from kserve.models.v1beta1_inference_services_config import V1beta1InferenceServicesConfig
from kserve.models.v1beta1_ingress_config import V1beta1IngressConfig
from kserve.models.v1beta1_light_gbm_spec import V1beta1LightGBMSpec
from kserve.models.v1beta1_local_model_config import V1beta1LocalModelConfig
from kserve.models.v1beta1_logger_spec import V1beta1LoggerSpec
from kserve.models.v1beta1_model_copies import V1beta1ModelCopies
from kserve.models.v1beta1_model_format import V1beta1ModelFormat
from kserve.models.v1beta1_model_revision_states import V1beta1ModelRevisionStates
from kserve.models.v1beta1_model_spec import V1beta1ModelSpec
from kserve.models.v1beta1_model_status import V1beta1ModelStatus
from kserve.models.v1beta1_onnx_runtime_spec import V1beta1ONNXRuntimeSpec
from kserve.models.v1beta1_pmml_spec import V1beta1PMMLSpec
from kserve.models.v1beta1_paddle_server_spec import V1beta1PaddleServerSpec
from kserve.models.v1beta1_pod_spec import V1beta1PodSpec
from kserve.models.v1beta1_predictor_extension_spec import V1beta1PredictorExtensionSpec
from kserve.models.v1beta1_predictor_spec import V1beta1PredictorSpec
from kserve.models.v1beta1_resource_config import V1beta1ResourceConfig
from kserve.models.v1beta1_sk_learn_spec import V1beta1SKLearnSpec
from kserve.models.v1beta1_security_config import V1beta1SecurityConfig
from kserve.models.v1beta1_service_config import V1beta1ServiceConfig
from kserve.models.v1beta1_storage_spec import V1beta1StorageSpec
from kserve.models.v1beta1_tf_serving_spec import V1beta1TFServingSpec
from kserve.models.v1beta1_torch_serve_spec import V1beta1TorchServeSpec
from kserve.models.v1beta1_transformer_spec import V1beta1TransformerSpec
from kserve.models.v1beta1_triton_spec import V1beta1TritonSpec
from kserve.models.v1beta1_worker_spec import V1beta1WorkerSpec
from kserve.models.v1beta1_xg_boost_spec import V1beta1XGBoostSpec



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import glob
import gzip
import json
import mimetypes
import os
import re
import shutil
import tarfile
import tempfile
import time
import zipfile
from pathlib import Path
from typing import Dict
from urllib.parse import urlparse
import requests

from ..logging import logger

MODEL_MOUNT_DIRS = "/mnt/models"

_GCS_PREFIX = "gs://"
_S3_PREFIX = "s3://"
_HDFS_PREFIX = "hdfs://"
_WEBHDFS_PREFIX = "webhdfs://"
_AZURE_BLOB_RE = [
    "https://(.+?).blob.core.windows.net/(.+)",
    "https://(.+?).z[0-9]{1,2}.blob.storage.azure.net/(.+)",
]
_AZURE_FILE_RE = [
    "https://(.+?).file.core.windows.net/(.+)",
    "https://(.+?).z[0-9]{1,2}.file.storage.azure.net/(.+)",
]
_LOCAL_PREFIX = "file://"
_URI_RE = "https?://(.+)/(.+)"
_HTTP_PREFIX = "http(s)://"
_HEADERS_SUFFIX = "-headers"
_PVC_PREFIX = "/mnt/pvc"
_HF_PREFIX = "hf://"

_HDFS_SECRET_DIRECTORY = "/var/secrets/kserve-hdfscreds"
_HDFS_FILE_SECRETS = ["KERBEROS_KEYTAB", "TLS_CERT", "TLS_KEY", "TLS_CA"]


class Storage(object):
    @staticmethod
    def download(uri: str, out_dir: str = None) -> str:
        start = time.monotonic()
        Storage._update_with_storage_spec()
        logger.info("Copying contents of %s to local", uri)

        if uri.startswith(_PVC_PREFIX) and not os.path.exists(uri):
            raise Exception(f"Cannot locate source uri {uri} for PVC")

        is_local = uri.startswith(_LOCAL_PREFIX) or os.path.exists(uri)
        if is_local:
            if out_dir is None:
                # noop if out_dir is not set and the path is local
                model_dir = Storage._download_local(uri)
            else:
                if not os.path.exists(out_dir):
                    os.mkdir(out_dir)
                model_dir = Storage._download_local(uri, out_dir)
        else:
            if out_dir is None:
                out_dir = tempfile.mkdtemp()
            elif not os.path.exists(out_dir):
                os.mkdir(out_dir)

            if uri.startswith(MODEL_MOUNT_DIRS):
                # Don't need to download models if this InferenceService is running in the multi-model
                # serving mode. The model agent will download models.
                model_dir = out_dir
            elif uri.startswith(_GCS_PREFIX):
                model_dir = Storage._download_gcs(uri, out_dir)
            elif uri.startswith(_S3_PREFIX):
                model_dir = Storage._download_s3(uri, out_dir)
            elif uri.startswith(_HDFS_PREFIX) or uri.startswith(_WEBHDFS_PREFIX):
                model_dir = Storage._download_hdfs(uri, out_dir)
            elif any(re.search(pattern, uri) for pattern in _AZURE_BLOB_RE):
                model_dir = Storage._download_azure_blob(uri, out_dir)
            elif any(re.search(pattern, uri) for pattern in _AZURE_FILE_RE):
                model_dir = Storage._download_azure_file_share(uri, out_dir)
            elif re.search(_URI_RE, uri):
                model_dir = Storage._download_from_uri(uri, out_dir)
            elif uri.startswith(_HF_PREFIX):
                model_dir = Storage._download_hf(uri, out_dir)
            else:
                raise Exception(
                    "Cannot recognize storage type for "
                    + uri
                    + "\n'%s', '%s', '%s', and '%s' are the current available storage type."
                    % (_GCS_PREFIX, _S3_PREFIX, _LOCAL_PREFIX, _HTTP_PREFIX)
                )

        logger.info("Successfully copied %s to %s", uri, out_dir)
        logger.info(f"Model downloaded in {time.monotonic() - start} seconds.")
        return model_dir

    @staticmethod
    def _update_with_storage_spec():
        storage_secret_json = json.loads(os.environ.get("STORAGE_CONFIG", "{}"))
        storage_secret_override_params = json.loads(
            os.environ.get("STORAGE_OVERRIDE_CONFIG", "{}")
        )
        if storage_secret_override_params:
            for key, value in storage_secret_override_params.items():
                storage_secret_json[key] = value

        if storage_secret_json.get("type", "") == "s3":
            for env_var, key in (
                ("AWS_ENDPOINT_URL", "endpoint_url"),
                ("AWS_ACCESS_KEY_ID", "access_key_id"),
                ("AWS_SECRET_ACCESS_KEY", "secret_access_key"),
                ("AWS_DEFAULT_REGION", "region"),
                ("AWS_CA_BUNDLE", "ca_bundle"),
                ("S3_VERIFY_SSL", "verify_ssl"),
                ("awsAnonymousCredential", "anonymous"),
            ):
                if key in storage_secret_json:
                    os.environ[env_var] = storage_secret_json.get(key)

        if (
            storage_secret_json.get("type", "") == "hdfs"
            or storage_secret_json.get("type", "") == "webhdfs"
        ):
            temp_dir = tempfile.mkdtemp()
            os.environ["HDFS_SECRET_DIR"] = temp_dir
            for key, value in storage_secret_json.items():
                mode = "w"

                # If the secret is supposed to be a file, then it was base64 encoded in the json
                if key in _HDFS_FILE_SECRETS:
                    value = base64.b64decode(value)
                    mode = "wb"

                with open(f"{temp_dir}/{key}", mode) as f:
                    f.write(value)
                    f.flush()

    @staticmethod
    def get_S3_config():
        from botocore import UNSIGNED
        from botocore.client import Config

        # default s3 config
        c = Config()

        # anon environment variable defined in s3_secret.go
        anon = "true" == os.getenv("awsAnonymousCredential", "false").lower()
        # S3UseVirtualBucket environment variable defined in s3_secret.go
        # use virtual hosted-style URLs if enabled
        virtual = "true" == os.getenv("S3_USER_VIRTUAL_BUCKET", "false").lower()
        # S3UseAccelerate environment variable defined in s3_secret.go
        # use transfer acceleration if enabled
        accelerate = "true" == os.getenv("S3_USE_ACCELERATE", "false").lower()

        if anon:
            c = c.merge(Config(signature_version=UNSIGNED))
        if virtual:
            c = c.merge(Config(s3={"addressing_style": "virtual"}))
        if accelerate:
            c = c.merge(Config(s3={"use_accelerate_endpoint": accelerate}))

        # NOTE: If endpoint_url provided is legacy ("https://s3.amazonaws.com") and region is not global (us-east-1), set to virtual addressing style
        # So that request would not return PermanentRedirect due to region not in the endpoint url
        # AWS SDK retries under the hood to set the correct region when the valid virtual addressing style endpoint url is provided
        endpoint_url = os.getenv("AWS_ENDPOINT_URL")
        region = os.getenv("AWS_DEFAULT_REGION")
        if endpoint_url == "https://s3.amazonaws.com" and region not in (
            None,
            "us-east-1",
        ):
            c = c.merge(Config(s3={"addressing_style": "virtual"}))

        return c

    @staticmethod
    def _download_s3(uri, temp_dir: str) -> str:
        import boto3

        # Boto3 looks at various configuration locations until it finds configuration values.
        # lookup order:
        # 1. Config object passed in as the config parameter when creating S3 resource
        #    if awsAnonymousCredential env var true, passed in via config
        # 2. Environment variables
        # 3. ~/.aws/config file
        kwargs = {"config": Storage.get_S3_config()}
        endpoint_url = os.getenv("AWS_ENDPOINT_URL")
        if endpoint_url:
            kwargs.update({"endpoint_url": endpoint_url})
        verify_ssl = os.getenv("S3_VERIFY_SSL")
        if verify_ssl:
            verify_ssl = not verify_ssl.lower() in ["0", "false"]
            kwargs.update({"verify": verify_ssl})
        else:
            verify_ssl = True

        # If verify_ssl is true, then check there is custom ca bundle cert
        if verify_ssl:
            global_ca_bundle_configmap = os.getenv("CA_BUNDLE_CONFIGMAP_NAME")
            if global_ca_bundle_configmap:
                isvc_aws_ca_bundle_path = os.getenv("AWS_CA_BUNDLE")
                if isvc_aws_ca_bundle_path and isvc_aws_ca_bundle_path != "":
                    ca_bundle_full_path = isvc_aws_ca_bundle_path
                else:
                    global_ca_bundle_volume_mount_path = os.getenv(
                        "CA_BUNDLE_VOLUME_MOUNT_POINT"
                    )
                    ca_bundle_full_path = (
                        global_ca_bundle_volume_mount_path + "/cabundle.crt"
                    )
                if os.path.exists(ca_bundle_full_path):
                    logger.info("ca bundle file(%s) exists." % (ca_bundle_full_path))
                    kwargs.update({"verify": ca_bundle_full_path})
                else:
                    raise RuntimeError(
                        "Failed to find ca bundle file(%s)." % ca_bundle_full_path
                    )
        s3 = boto3.resource("s3", **kwargs)
        parsed = urlparse(uri, scheme="s3")
        bucket_name = parsed.netloc
        bucket_path = parsed.path.lstrip("/")

        file_count = 0
        exact_obj_found = False
        bucket = s3.Bucket(bucket_name)
        for obj in bucket.objects.filter(Prefix=bucket_path):
            # Skip where boto3 lists the directory as an object
            if obj.key.endswith("/"):
                continue
            # In the case where bucket_path points to a single object, set the target key to bucket_path
            # Otherwise, remove the bucket_path prefix, strip any extra slashes, then prepend the target_dir
            # Example:
            # s3://test-bucket
            # Objects: /a/b/c/model.bin /a/model.bin /model.bin
            #
            # If 'uri' is set to "s3://test-bucket", then the downloader will
            # download all the objects listed above, re-creating their subpaths
            # under the temp_dir.
            # If 'uri' is set to "s3://test-bucket/a", then the downloader will
            # add to temp_dir: b/c/model.bin and model.bin.
            # If 'uri' is set to "s3://test-bucket/a/b/c/model.bin", then
            # the downloader will add to temp dir: model.bin
            # (without any subpaths).
            # If the bucket path is s3://test/models
            # Objects: churn, churn-pickle, churn-pickle-logs
            bucket_path_last_part = bucket_path.split("/")[-1]
            object_last_path = obj.key.split("/")[-1]

            if bucket_path == obj.key:
                target_key = obj.key.rsplit("/", 1)[-1]
                exact_obj_found = True
            elif bucket_path_last_part and object_last_path.startswith(
                bucket_path_last_part
            ):
                target_key = object_last_path
            elif obj.key.startswith(bucket_path):
                # remove only the bucket_path prefix
                target_key = obj.key[len(bucket_path) :].lstrip("/")
            else:
                target_key = obj.key

            target = f"{temp_dir}/{target_key}"
            if not os.path.exists(os.path.dirname(target)):
                os.makedirs(os.path.dirname(target), exist_ok=True)
            bucket.download_file(obj.key, target)
            logger.info("Downloaded object %s to %s" % (obj.key, target))
            file_count += 1

            # If the exact object is found, then it is sufficient to download that and break the loop
            if exact_obj_found:
                break
        if file_count == 0:
            raise RuntimeError(
                "Failed to fetch model. No model found in %s." % bucket_path
            )

        # Unpack compressed file, supports .tgz, tar.gz and zip file formats.
        if file_count == 1:
            mimetype, _ = mimetypes.guess_type(target)
            if mimetype in ["application/x-tar", "application/zip"]:
                temp_dir = Storage._unpack_archive_file(target, mimetype, temp_dir)
        return temp_dir

    @staticmethod
    def _download_hf(uri, temp_dir: str) -> str:
        from huggingface_hub import snapshot_download

        components = uri[len(_HF_PREFIX) :].split("/")

        # Validate that the URI has two parts: repo and model (optional hash)
        if len(components) != 2:
            raise ValueError(
                "URI must contain exactly one '/' separating the repo and model name"
            )

        repo = components[0]
        model_part = components[1]

        if not repo:
            raise ValueError("Repository name cannot be empty")
        if not model_part:
            raise ValueError("Model name cannot be empty")

        model, _, hash_value = model_part.partition(":")
        # Ensure model is non-empty
        if not model:
            raise ValueError("Model name cannot be empty")

        revision = hash_value if hash_value else None

        snapshot_download(
            repo_id=f"{repo}/{model}", revision=revision, local_dir=temp_dir
        )
        return temp_dir

    @staticmethod
    def _download_gcs(uri, temp_dir: str) -> str:
        from google.auth import exceptions
        from google.cloud import storage
        import copy

        try:
            storage_client = storage.Client()
        except exceptions.DefaultCredentialsError:
            storage_client = storage.Client.create_anonymous_client()
        bucket_args = uri.replace(_GCS_PREFIX, "", 1).split("/", 1)
        bucket_name = bucket_args[0]
        bucket_path = bucket_args[1] if len(bucket_args) > 1 else ""
        bucket = storage_client.bucket(bucket_name)
        prefix = bucket_path
        if not prefix.endswith("/"):
            prefix = prefix + "/"
        blobs = bucket.list_blobs(prefix=prefix)
        file_count = 0

        # Shallow copy, otherwise Iterator has already started
        shallow_blobs = copy.copy(blobs)
        blob = bucket.blob(bucket_path)
        # checks if the blob is a file or a directory
        if blob.name == bucket_path and len(list(shallow_blobs)) == 0:
            dest_path = os.path.join(temp_dir, os.path.basename(bucket_path))
            logger.info("Downloading single file to: %s", dest_path)
            blob.download_to_filename(dest_path)
            file_count = 1

        else:
            for blob in blobs:
                # Replace any prefix from the object key with temp_dir
                subdir_object_key = blob.name.replace(bucket_path, "", 1).lstrip("/")
                # Create necessary subdirectory to store the object locally
                if "/" in subdir_object_key:
                    local_object_dir = os.path.join(
                        temp_dir, subdir_object_key.rsplit("/", 1)[0]
                    )
                    if not os.path.isdir(local_object_dir):
                        os.makedirs(local_object_dir, exist_ok=True)
                if subdir_object_key.strip() != "" and not subdir_object_key.endswith(
                    "/"
                ):
                    dest_path = os.path.join(temp_dir, subdir_object_key)
                    logger.info("Downloading: %s", dest_path)
                    blob.download_to_filename(dest_path)
                    file_count += 1

        if file_count == 0:
            raise RuntimeError("Failed to fetch model. No model found in %s." % uri)

        # Unpack compressed file, supports .tgz, tar.gz and zip file formats.
        if file_count == 1:
            mimetype, _ = mimetypes.guess_type(blob.name)
            if mimetype in ["application/x-tar", "application/zip"]:
                temp_dir = Storage._unpack_archive_file(dest_path, mimetype, temp_dir)
        return temp_dir

    @staticmethod
    def _load_hdfs_configuration() -> Dict:
        config = {
            "HDFS_NAMENODE": None,
            "USER_PROXY": None,
            "HDFS_ROOTPATH": None,
            "TLS_CERT": None,
            "TLS_KEY": None,
            "TLS_CA": None,
            "TLS_SKIP_VERIFY": "false",
            "HEADERS": None,
            "N_THREADS": "2",
            "KERBEROS_KEYTAB": None,
            "KERBEROS_PRINCIPAL": None,
        }

        secret_dir = _HDFS_SECRET_DIRECTORY
        if os.environ.get("HDFS_SECRET_DIR"):
            secret_dir = os.environ["HDFS_SECRET_DIR"]

        for filename in os.listdir(secret_dir):
            if filename not in config:
                continue

            # We don't read files which are supposed to be files, just save their path
            if filename in _HDFS_FILE_SECRETS:
                config[filename] = f"{secret_dir}/{filename}"
                continue

            # Read file and save value in config dict
            with open(f"{secret_dir}/{filename}") as f:
                config[filename] = f.read()

        return config

    @staticmethod
    def _download_hdfs(uri, out_dir: str) -> str:
        from krbcontext.context import krbContext
        from hdfs.ext.kerberos import Client, KerberosClient

        config = Storage._load_hdfs_configuration()

        logger.info(f"Using the following hdfs config\n{config}")

        # Remove hdfs:// or webhdfs:// from the uri to get just the path
        # e.g. hdfs://user/me/model -> user/me/model
        if uri.startswith(_HDFS_PREFIX):
            path = uri[len(_HDFS_PREFIX) :]
        else:
            path = uri[len(_WEBHDFS_PREFIX) :]

        if not config["HDFS_ROOTPATH"]:
            path = "/" + path

        s = requests.Session()

        if config["TLS_CERT"]:
            s.cert = (config["TLS_CERT"], config["TLS_KEY"])
        # s.verify = , True, False, or CA PATH
        if config["TLS_CA"]:
            s.verify = config["TLS_CA"]
        if config["TLS_SKIP_VERIFY"].lower() == "true":
            s.verify = False

        if config["HEADERS"]:
            headers = json.loads(config["HEADERS"])
            s.headers.update(headers)

        if config["KERBEROS_PRINCIPAL"]:
            context = krbContext(
                using_keytab=True,
                principal=config["KERBEROS_PRINCIPAL"],
                keytab_file=config["KERBEROS_KEYTAB"],
            )
            context.init_with_keytab()
            client = KerberosClient(
                config["HDFS_NAMENODE"],
                proxy=config["USER_PROXY"],
                root=config["HDFS_ROOTPATH"],
                session=s,
            )
        else:
            client = Client(
                config["HDFS_NAMENODE"],
                proxy=config["USER_PROXY"],
                root=config["HDFS_ROOTPATH"],
                session=s,
            )
        file_count = 0
        dest_file_path = ""

        # Check path exists and get path status
        # Raises HdfsError when path does not exist
        status = client.status(path)

        if status["type"] == "FILE":
            client.download(path, out_dir, n_threads=1)
            file_count += 1
            file_name = path.rsplit("/", 1)[-1]
            dest_file_path = f"{out_dir}/{file_name}"
        else:
            files = client.list(path)
            file_count += len(files)
            for f in files:
                client.download(
                    f"{path}/{f}", out_dir, n_threads=int(config["N_THREADS"])
                )
                dest_file_path = f"{out_dir}/{f}"

        if file_count == 1:
            mimetype, _ = mimetypes.guess_type(dest_file_path)
            if mimetype in ["application/x-tar", "application/zip"]:
                out_dir = Storage._unpack_archive_file(
                    dest_file_path, mimetype, out_dir
                )
        return out_dir

    @staticmethod
    def _download_azure_blob(
        uri, out_dir: str
    ) -> str:  # pylint: disable=too-many-locals
        from azure.storage.blob import BlobServiceClient
        from azure.storage.blob._list_blobs_helper import BlobPrefix

        account_name, account_url, container_name, prefix = Storage._parse_azure_uri(
            uri
        )
        logger.info(
            "Connecting to BLOB account: [%s], container: [%s], prefix: [%s]",
            account_name,
            container_name,
            prefix,
        )
        token = (
            Storage._get_azure_storage_token()
            or Storage._get_azure_storage_access_key()
        )
        if token is None:
            logger.warning(
                "Azure credentials or shared access signature token not found, retrying anonymous access"
            )

        blob_service_client = BlobServiceClient(account_url, credential=token)
        container_client = blob_service_client.get_container_client(container_name)
        file_count = 0
        blobs = []
        max_depth = 5
        stack = [(prefix, max_depth)]
        while stack:
            curr_prefix, depth = stack.pop()
            if depth < 0:
                continue
            for item in container_client.walk_blobs(name_starts_with=curr_prefix):
                if isinstance(item, BlobPrefix):
                    stack.append((item.name, depth - 1))
                else:
                    blobs += container_client.list_blobs(
                        name_starts_with=item.name, include=["snapshots"]
                    )
        for blob in blobs:
            file_name = blob.name.replace(prefix, "", 1).lstrip("/")
            if not file_name:
                file_name = os.path.basename(prefix)
            dest_path = os.path.join(out_dir, file_name)
            Path(os.path.dirname(dest_path)).mkdir(parents=True, exist_ok=True)
            logger.info("Downloading: %s to %s", blob.name, dest_path)
            downloader = container_client.download_blob(blob.name)
            with open(dest_path, "wb+") as f:
                f.write(downloader.readall())
            file_count += 1
        if file_count == 0:
            raise RuntimeError("Failed to fetch model. No model found in %s." % (uri))

        # Unpack compressed file, supports .tgz, tar.gz and zip file formats.
        if file_count == 1:
            mimetype, _ = mimetypes.guess_type(dest_path)
            if mimetype in ["application/x-tar", "application/zip"]:
                out_dir = Storage._unpack_archive_file(dest_path, mimetype, out_dir)
        return out_dir

    @staticmethod
    def _download_azure_file_share(
        uri, out_dir: str
    ) -> str:  # pylint: disable=too-many-locals
        from azure.storage.fileshare import ShareServiceClient

        account_name, account_url, share_name, prefix = Storage._parse_azure_uri(uri)
        logger.info(
            "Connecting to file share account: [%s], container: [%s], prefix: [%s]",
            account_name,
            share_name,
            prefix,
        )
        access_key = Storage._get_azure_storage_access_key()
        if access_key is None:
            logger.warning(
                "Azure storage access key not found, retrying anonymous access"
            )

        share_service_client = ShareServiceClient(account_url, credential=access_key)
        share_client = share_service_client.get_share_client(share_name)
        file_count = 0
        share_files = []
        max_depth = 5
        stack = [(prefix, max_depth)]
        while stack:
            curr_prefix, depth = stack.pop()
            if depth < 0:
                continue
            for item in share_client.list_directories_and_files(
                directory_name=curr_prefix
            ):
                if item.is_directory:
                    stack.append(
                        ("/".join([curr_prefix, item.name]).strip("/"), depth - 1)
                    )
                else:
                    share_files.append((curr_prefix, item))
        for prefix, file_item in share_files:
            parts = [prefix] if prefix else []
            parts.append(file_item.name)
            file_path = "/".join(parts).lstrip("/")
            dest_path = os.path.join(out_dir, file_path)
            Path(os.path.dirname(dest_path)).mkdir(parents=True, exist_ok=True)
            logger.info("Downloading: %s to %s", file_item.name, dest_path)
            file_client = share_client.get_file_client(file_path)
            with open(dest_path, "wb+") as f:
                data = file_client.download_file()
                data.readinto(f)
            file_count += 1
        if file_count == 0:
            raise RuntimeError("Failed to fetch model. No model found in %s." % (uri))

        # Unpack compressed file, supports .tgz, tar.gz and zip file formats.
        if file_count == 1:
            mimetype, _ = mimetypes.guess_type(dest_path)
            if mimetype in ["application/x-tar", "application/zip"]:
                out_dir = Storage._unpack_archive_file(dest_path, mimetype, out_dir)
        return out_dir

    @staticmethod
    def _parse_azure_uri(uri):  # pylint: disable=too-many-locals
        parsed = urlparse(uri)
        account_name = parsed.netloc.split(".")[0]
        account_url = "https://{}{}".format(
            parsed.netloc, "?" + parsed.query if parsed.query else ""
        )
        object_name, prefix = parsed.path.lstrip("/").split("/", 1)
        prefix = prefix.strip("/")
        return account_name, account_url, object_name, prefix

    @staticmethod
    def _get_azure_storage_token():
        tenant_id = os.getenv("AZ_TENANT_ID", "")
        client_id = os.getenv("AZ_CLIENT_ID", "")
        client_secret = os.getenv("AZ_CLIENT_SECRET", "")

        # convert old environment variable to conform Azure defaults
        # see azure/identity/_constants.py
        if tenant_id:
            os.environ["AZURE_TENANT_ID"] = tenant_id
        if client_id:
            os.environ["AZURE_CLIENT_ID"] = client_id
        if client_secret:
            os.environ["AZURE_CLIENT_SECRET"] = client_secret

        client_id = os.getenv("AZURE_CLIENT_ID", "")
        if not client_id:
            return None

        # note the SP must have "Storage Blob Data Owner" perms for this to work
        from azure.identity import DefaultAzureCredential

        token_credential = DefaultAzureCredential()

        logger.info("Retrieved SP token credential for client_id: %s", client_id)
        return token_credential

    @staticmethod
    def _get_azure_storage_access_key():
        return os.getenv("AZURE_STORAGE_ACCESS_KEY")

    @staticmethod
    def _download_local(uri, out_dir=None) -> str:
        local_path = uri.replace(_LOCAL_PREFIX, "", 1)
        if not os.path.exists(local_path):
            raise RuntimeError("Local path %s does not exist." % (uri))

        if out_dir is None:
            return local_path
        elif not os.path.isdir(out_dir):
            os.makedirs(out_dir, exist_ok=True)

        if os.path.isdir(local_path):
            local_path = os.path.join(local_path, "*")

        file_count = 0
        for src in glob.glob(local_path):
            _, tail = os.path.split(src)
            dest_path = os.path.join(out_dir, tail)
            logger.info("Linking: %s to %s", src, dest_path)
            if not os.path.exists(dest_path):
                os.symlink(src, dest_path)
            else:
                logger.info("File %s already exist", dest_path)
            file_count += 1
        if file_count == 0:
            raise RuntimeError("Failed to fetch model. No model found in %s." % (uri))
        # Unpack compressed file, supports .tgz, tar.gz and zip file formats.
        if file_count == 1:
            mimetype, _ = mimetypes.guess_type(dest_path)
            if mimetype in ["application/x-tar", "application/zip"]:
                out_dir = Storage._unpack_archive_file(dest_path, mimetype, out_dir)
        return out_dir

    @staticmethod
    def _download_from_uri(uri, out_dir=None) -> str:
        url = urlparse(uri)
        filename = os.path.basename(url.path)
        # Determine if the symbol '?' exists in the path
        if mimetypes.guess_type(url.path)[0] is None and url.query != "":
            mimetype, encoding = mimetypes.guess_type(url.query)
        else:
            mimetype, encoding = mimetypes.guess_type(url.path)
        local_path = os.path.join(out_dir, filename)

        if filename == "":
            raise ValueError("No filename contained in URI: %s" % (uri))

        # Get header information from host url
        headers = {}
        host_uri = url.hostname

        headers_json = os.getenv(host_uri + _HEADERS_SUFFIX, "{}")
        headers = json.loads(headers_json)

        with requests.get(uri, stream=True, headers=headers) as response:
            if response.status_code != 200:
                raise RuntimeError(
                    "URI: %s returned a %s response code." % (uri, response.status_code)
                )
            zip_content_types = (
                "application/x-zip-compressed",
                "application/zip",
                "application/zip-compressed",
            )
            if mimetype == "application/zip" and not response.headers.get(
                "Content-Type", ""
            ).startswith(zip_content_types):
                raise RuntimeError(
                    "URI: %s did not respond with any of following 'Content-Type': "
                    % uri
                    + ", ".join(zip_content_types)
                )
            tar_content_types = (
                "application/x-tar",
                "application/x-gtar",
                "application/x-gzip",
                "application/gzip",
            )
            if mimetype == "application/x-tar" and not response.headers.get(
                "Content-Type", ""
            ).startswith(tar_content_types):
                raise RuntimeError(
                    "URI: %s did not respond with any of following 'Content-Type': "
                    % uri
                    + ", ".join(tar_content_types)
                )
            if (
                mimetype != "application/zip" and mimetype != "application/x-tar"
            ) and not response.headers.get("Content-Type", "").startswith(
                "application/octet-stream"
            ):
                raise RuntimeError(
                    "URI: %s did not respond with 'Content-Type': 'application/octet-stream'"
                    % uri
                )

            if encoding == "gzip":
                stream = gzip.GzipFile(fileobj=response.raw)
                local_path = os.path.join(out_dir, f"{filename}.tar")
            else:
                stream = response.raw
            with open(local_path, "wb") as out:
                shutil.copyfileobj(stream, out)

        if mimetype in ["application/x-tar", "application/zip"]:
            out_dir = Storage._unpack_archive_file(local_path, mimetype, out_dir)
        return out_dir

    @staticmethod
    def _unpack_archive_file(file_path, mimetype, target_dir=None) -> str:
        if not target_dir:
            target_dir = os.path.dirname(file_path)

        try:
            logger.info("Unpacking: %s", file_path)
            if mimetype == "application/x-tar":
                archive = tarfile.open(file_path, "r", encoding="utf-8")
            else:
                archive = zipfile.ZipFile(file_path, "r")
            archive.extractall(target_dir)
            archive.close()
        except (tarfile.TarError, zipfile.BadZipfile):
            raise RuntimeError(
                "Failed to unpack archive file. \
The file format is not valid."
            )
        os.remove(file_path)
        return target_dir



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
    KServe

    Python SDK for KServe  # noqa: E501

    OpenAPI spec version: v0.1

    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six

from kserve.models.knative_condition import KnativeCondition  # noqa: F401,E501


class KnativeStatus(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'conditions': 'list[KnativeCondition]',
        'observed_generation': 'int'
    }

    attribute_map = {
        'conditions': 'conditions',
        'observed_generation': 'observedGeneration'
    }

    def __init__(self, conditions=None, observed_generation=None):  # noqa: E501
        """KnativeStatus - a model defined in Swagger"""  # noqa: E501

        self._conditions = None
        self._observed_generation = None
        self.discriminator = None

        if conditions is not None:
            self.conditions = conditions
        if observed_generation is not None:
            self.observed_generation = observed_generation

    @property
    def conditions(self):
        """Gets the conditions of this KnativeStatus.  # noqa: E501

        Conditions the latest available observations of a resource's current state.  # noqa: E501

        :return: The conditions of this KnativeStatus.  # noqa: E501
        :rtype: list[KnativeCondition]
        """
        return self._conditions

    @conditions.setter
    def conditions(self, conditions):
        """Sets the conditions of this KnativeStatus.

        Conditions the latest available observations of a resource's current state.  # noqa: E501

        :param conditions: The conditions of this KnativeStatus.  # noqa: E501
        :type: list[KnativeCondition]
        """

        self._conditions = conditions

    @property
    def observed_generation(self):
        """Gets the observed_generation of this KnativeStatus.  # noqa: E501

        ObservedGeneration is the 'Generation' of the Service that was last processed by the controller.  # noqa: E501

        :return: The observed_generation of this KnativeStatus.  # noqa: E501
        :rtype: int
        """
        return self._observed_generation

    @observed_generation.setter
    def observed_generation(self, observed_generation):
        """Sets the observed_generation of this KnativeStatus.

        ObservedGeneration is the 'Generation' of the Service that was last processed by the controller.  # noqa: E501

        :param observed_generation: The observed_generation of this KnativeStatus.  # noqa: E501
        :type: int
        """

        self._observed_generation = observed_generation

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(KnativeStatus, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, KnativeStatus):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Awaitable, Callable

from grpc import HandlerCallDetails, RpcMethodHandler, ServicerContext, StatusCode
from grpc.aio import ServerInterceptor
from grpc_interceptor import AsyncExceptionToStatusInterceptor
from grpc_interceptor.exceptions import GrpcException

from ...errors import (
    InferenceError,
    InvalidInput,
    ModelNotFound,
    ModelNotReady,
    UnsupportedProtocol,
    ServerNotReady,
    ServerNotLive,
)
from ...logging import logger


class LoggingInterceptor(ServerInterceptor):

    async def intercept_service(
        self,
        continuation: Callable[[HandlerCallDetails], Awaitable[RpcMethodHandler]],
        handler_call_details: HandlerCallDetails,
    ) -> RpcMethodHandler:
        logger.info(f"grpc method: {handler_call_details.method}")
        return await continuation(handler_call_details)


class ExceptionToStatusInterceptor(AsyncExceptionToStatusInterceptor):

    async def handle_exception(
        self,
        ex: Exception,
        request_or_iterator: Any,
        context: ServicerContext,
        method_name: str,
    ):
        if isinstance(ex, (InferenceError, InvalidInput, UnsupportedProtocol)):
            logger.error(
                "Error processing inference request: %s, grpc_method: %s",
                ex,
                method_name,
                exc_info=True,
            )
            await context.abort(StatusCode.INVALID_ARGUMENT, ex.reason)
        elif isinstance(ex, (ServerNotReady, ServerNotLive)):
            logger.error(
                "Unable to process the request: %s, grpc_method: %s",
                ex,
                method_name,
                exc_info=True,
            )
            await context.abort(StatusCode.UNAVAILABLE, ex.reason)
        elif isinstance(ex, NotImplementedError):
            logger.error(
                "Method not implemented: %s, grpc_method: %s",
                ex,
                method_name,
                exc_info=True,
            )
            await context.abort(StatusCode.UNIMPLEMENTED, ex.reason)
        elif isinstance(ex, ModelNotFound):
            logger.error(
                "Model not found: %s, grpc_method: %s", ex, method_name, exc_info=True
            )
            await context.abort(StatusCode.NOT_FOUND, ex.reason)
        elif isinstance(ex, ModelNotReady):
            logger.error(
                "Model not ready: %s, grpc_method: %s", ex, method_name, exc_info=True
            )
            await context.abort(StatusCode.UNAVAILABLE, ex.reason)
        elif isinstance(ex, GrpcException):
            await super().handle_exception(
                ex, request_or_iterator, context, method_name
            )
        else:
            logger.error(
                "Internal server error: %s, grpc_method: %s",
                ex,
                method_name,
                exc_info=True,
            )
            await context.abort(StatusCode.INTERNAL, "Internal server error")



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from kubernetes import client
from kubernetes import watch as k8s_watch
from tabulate import tabulate

from ..constants import constants
from ..utils import utils


def isvc_watch(name=None, namespace=None, timeout_seconds=600, generation=0):
    """Watch the created or patched InferenceService in the specified namespace"""

    if namespace is None:
        namespace = utils.get_default_target_namespace()

    headers = ["NAME", "READY", "PREV", "LATEST", "URL"]
    table_fmt = "plain"

    stream = k8s_watch.Watch().stream(
        client.CustomObjectsApi().list_namespaced_custom_object,
        constants.KSERVE_GROUP,
        constants.KSERVE_V1BETA1_VERSION,
        namespace,
        constants.KSERVE_PLURAL_INFERENCESERVICE,
        timeout_seconds=timeout_seconds,
    )

    for event in stream:
        isvc = event["object"]
        isvc_name = isvc["metadata"]["name"]
        if name and name != isvc_name:
            continue
        else:
            status = "Unknown"
            if isvc.get("status", ""):
                url = isvc["status"].get("url", "")
                traffic = (
                    isvc["status"]
                    .get("components", {})
                    .get("predictor", {})
                    .get("traffic", [])
                )
                traffic_percent = 100
                if constants.OBSERVED_GENERATION in isvc["status"]:
                    observed_generation = isvc["status"][constants.OBSERVED_GENERATION]
                    for t in traffic:
                        if t["latestRevision"]:
                            traffic_percent = t["percent"]

                    if generation != 0 and observed_generation != generation:
                        continue
                    for condition in isvc["status"].get("conditions", {}):
                        if condition.get("type", "") == "Ready":
                            status = condition.get("status", "Unknown")
                    print(
                        tabulate(
                            [
                                [
                                    isvc_name,
                                    status,
                                    100 - traffic_percent,
                                    traffic_percent,
                                    url,
                                ]
                            ],
                            headers=headers,
                            tablefmt=table_fmt,
                        )
                    )
                    if status == "True":
                        break

            else:
                print(
                    tabulate(
                        [[isvc_name, status, "", "", ""]],
                        headers=headers,
                        tablefmt=table_fmt,
                    )
                )
                # Sleep 2 to avoid status section is not generated within a very short time.
                time.sleep(2)
                continue



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class InferParameter(_message.Message):
    __slots__ = ["bool_param", "int64_param", "string_param"]
    BOOL_PARAM_FIELD_NUMBER: _ClassVar[int]
    INT64_PARAM_FIELD_NUMBER: _ClassVar[int]
    STRING_PARAM_FIELD_NUMBER: _ClassVar[int]
    bool_param: bool
    int64_param: int
    string_param: str
    def __init__(self, bool_param: bool = ..., int64_param: _Optional[int] = ..., string_param: _Optional[str] = ...) -> None: ...

class InferTensorContents(_message.Message):
    __slots__ = ["bool_contents", "bytes_contents", "fp32_contents", "fp64_contents", "int64_contents", "int_contents", "uint64_contents", "uint_contents"]
    BOOL_CONTENTS_FIELD_NUMBER: _ClassVar[int]
    BYTES_CONTENTS_FIELD_NUMBER: _ClassVar[int]
    FP32_CONTENTS_FIELD_NUMBER: _ClassVar[int]
    FP64_CONTENTS_FIELD_NUMBER: _ClassVar[int]
    INT64_CONTENTS_FIELD_NUMBER: _ClassVar[int]
    INT_CONTENTS_FIELD_NUMBER: _ClassVar[int]
    UINT64_CONTENTS_FIELD_NUMBER: _ClassVar[int]
    UINT_CONTENTS_FIELD_NUMBER: _ClassVar[int]
    bool_contents: _containers.RepeatedScalarFieldContainer[bool]
    bytes_contents: _containers.RepeatedScalarFieldContainer[bytes]
    fp32_contents: _containers.RepeatedScalarFieldContainer[float]
    fp64_contents: _containers.RepeatedScalarFieldContainer[float]
    int64_contents: _containers.RepeatedScalarFieldContainer[int]
    int_contents: _containers.RepeatedScalarFieldContainer[int]
    uint64_contents: _containers.RepeatedScalarFieldContainer[int]
    uint_contents: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, bool_contents: _Optional[_Iterable[bool]] = ..., int_contents: _Optional[_Iterable[int]] = ..., int64_contents: _Optional[_Iterable[int]] = ..., uint_contents: _Optional[_Iterable[int]] = ..., uint64_contents: _Optional[_Iterable[int]] = ..., fp32_contents: _Optional[_Iterable[float]] = ..., fp64_contents: _Optional[_Iterable[float]] = ..., bytes_contents: _Optional[_Iterable[bytes]] = ...) -> None: ...

class ModelInferRequest(_message.Message):
    __slots__ = ["id", "inputs", "model_name", "model_version", "outputs", "parameters", "raw_input_contents"]
    class InferInputTensor(_message.Message):
        __slots__ = ["contents", "datatype", "name", "parameters", "shape"]
        class ParametersEntry(_message.Message):
            __slots__ = ["key", "value"]
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: InferParameter
            def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[InferParameter, _Mapping]] = ...) -> None: ...
        CONTENTS_FIELD_NUMBER: _ClassVar[int]
        DATATYPE_FIELD_NUMBER: _ClassVar[int]
        NAME_FIELD_NUMBER: _ClassVar[int]
        PARAMETERS_FIELD_NUMBER: _ClassVar[int]
        SHAPE_FIELD_NUMBER: _ClassVar[int]
        contents: InferTensorContents
        datatype: str
        name: str
        parameters: _containers.MessageMap[str, InferParameter]
        shape: _containers.RepeatedScalarFieldContainer[int]
        def __init__(self, name: _Optional[str] = ..., datatype: _Optional[str] = ..., shape: _Optional[_Iterable[int]] = ..., parameters: _Optional[_Mapping[str, InferParameter]] = ..., contents: _Optional[_Union[InferTensorContents, _Mapping]] = ...) -> None: ...
    class InferRequestedOutputTensor(_message.Message):
        __slots__ = ["name", "parameters"]
        class ParametersEntry(_message.Message):
            __slots__ = ["key", "value"]
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: InferParameter
            def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[InferParameter, _Mapping]] = ...) -> None: ...
        NAME_FIELD_NUMBER: _ClassVar[int]
        PARAMETERS_FIELD_NUMBER: _ClassVar[int]
        name: str
        parameters: _containers.MessageMap[str, InferParameter]
        def __init__(self, name: _Optional[str] = ..., parameters: _Optional[_Mapping[str, InferParameter]] = ...) -> None: ...
    class ParametersEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: InferParameter
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[InferParameter, _Mapping]] = ...) -> None: ...
    ID_FIELD_NUMBER: _ClassVar[int]
    INPUTS_FIELD_NUMBER: _ClassVar[int]
    MODEL_NAME_FIELD_NUMBER: _ClassVar[int]
    MODEL_VERSION_FIELD_NUMBER: _ClassVar[int]
    OUTPUTS_FIELD_NUMBER: _ClassVar[int]
    PARAMETERS_FIELD_NUMBER: _ClassVar[int]
    RAW_INPUT_CONTENTS_FIELD_NUMBER: _ClassVar[int]
    id: str
    inputs: _containers.RepeatedCompositeFieldContainer[ModelInferRequest.InferInputTensor]
    model_name: str
    model_version: str
    outputs: _containers.RepeatedCompositeFieldContainer[ModelInferRequest.InferRequestedOutputTensor]
    parameters: _containers.MessageMap[str, InferParameter]
    raw_input_contents: _containers.RepeatedScalarFieldContainer[bytes]
    def __init__(self, model_name: _Optional[str] = ..., model_version: _Optional[str] = ..., id: _Optional[str] = ..., parameters: _Optional[_Mapping[str, InferParameter]] = ..., inputs: _Optional[_Iterable[_Union[ModelInferRequest.InferInputTensor, _Mapping]]] = ..., outputs: _Optional[_Iterable[_Union[ModelInferRequest.InferRequestedOutputTensor, _Mapping]]] = ..., raw_input_contents: _Optional[_Iterable[bytes]] = ...) -> None: ...

class ModelInferResponse(_message.Message):
    __slots__ = ["id", "model_name", "model_version", "outputs", "parameters", "raw_output_contents"]
    class InferOutputTensor(_message.Message):
        __slots__ = ["contents", "datatype", "name", "parameters", "shape"]
        class ParametersEntry(_message.Message):
            __slots__ = ["key", "value"]
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: InferParameter
            def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[InferParameter, _Mapping]] = ...) -> None: ...
        CONTENTS_FIELD_NUMBER: _ClassVar[int]
        DATATYPE_FIELD_NUMBER: _ClassVar[int]
        NAME_FIELD_NUMBER: _ClassVar[int]
        PARAMETERS_FIELD_NUMBER: _ClassVar[int]
        SHAPE_FIELD_NUMBER: _ClassVar[int]
        contents: InferTensorContents
        datatype: str
        name: str
        parameters: _containers.MessageMap[str, InferParameter]
        shape: _containers.RepeatedScalarFieldContainer[int]
        def __init__(self, name: _Optional[str] = ..., datatype: _Optional[str] = ..., shape: _Optional[_Iterable[int]] = ..., parameters: _Optional[_Mapping[str, InferParameter]] = ..., contents: _Optional[_Union[InferTensorContents, _Mapping]] = ...) -> None: ...
    class ParametersEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: InferParameter
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[InferParameter, _Mapping]] = ...) -> None: ...
    ID_FIELD_NUMBER: _ClassVar[int]
    MODEL_NAME_FIELD_NUMBER: _ClassVar[int]
    MODEL_VERSION_FIELD_NUMBER: _ClassVar[int]
    OUTPUTS_FIELD_NUMBER: _ClassVar[int]
    PARAMETERS_FIELD_NUMBER: _ClassVar[int]
    RAW_OUTPUT_CONTENTS_FIELD_NUMBER: _ClassVar[int]
    id: str
    model_name: str
    model_version: str
    outputs: _containers.RepeatedCompositeFieldContainer[ModelInferResponse.InferOutputTensor]
    parameters: _containers.MessageMap[str, InferParameter]
    raw_output_contents: _containers.RepeatedScalarFieldContainer[bytes]
    def __init__(self, model_name: _Optional[str] = ..., model_version: _Optional[str] = ..., id: _Optional[str] = ..., parameters: _Optional[_Mapping[str, InferParameter]] = ..., outputs: _Optional[_Iterable[_Union[ModelInferResponse.InferOutputTensor, _Mapping]]] = ..., raw_output_contents: _Optional[_Iterable[bytes]] = ...) -> None: ...

class ModelMetadataRequest(_message.Message):
    __slots__ = ["name", "version"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    name: str
    version: str
    def __init__(self, name: _Optional[str] = ..., version: _Optional[str] = ...) -> None: ...

class ModelMetadataResponse(_message.Message):
    __slots__ = ["inputs", "name", "outputs", "platform", "versions"]
    class TensorMetadata(_message.Message):
        __slots__ = ["datatype", "name", "shape"]
        DATATYPE_FIELD_NUMBER: _ClassVar[int]
        NAME_FIELD_NUMBER: _ClassVar[int]
        SHAPE_FIELD_NUMBER: _ClassVar[int]
        datatype: str
        name: str
        shape: _containers.RepeatedScalarFieldContainer[int]
        def __init__(self, name: _Optional[str] = ..., datatype: _Optional[str] = ..., shape: _Optional[_Iterable[int]] = ...) -> None: ...
    INPUTS_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    OUTPUTS_FIELD_NUMBER: _ClassVar[int]
    PLATFORM_FIELD_NUMBER: _ClassVar[int]
    VERSIONS_FIELD_NUMBER: _ClassVar[int]
    inputs: _containers.RepeatedCompositeFieldContainer[ModelMetadataResponse.TensorMetadata]
    name: str
    outputs: _containers.RepeatedCompositeFieldContainer[ModelMetadataResponse.TensorMetadata]
    platform: str
    versions: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, name: _Optional[str] = ..., versions: _Optional[_Iterable[str]] = ..., platform: _Optional[str] = ..., inputs: _Optional[_Iterable[_Union[ModelMetadataResponse.TensorMetadata, _Mapping]]] = ..., outputs: _Optional[_Iterable[_Union[ModelMetadataResponse.TensorMetadata, _Mapping]]] = ...) -> None: ...

class ModelReadyRequest(_message.Message):
    __slots__ = ["name", "version"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    name: str
    version: str
    def __init__(self, name: _Optional[str] = ..., version: _Optional[str] = ...) -> None: ...

class ModelReadyResponse(_message.Message):
    __slots__ = ["ready"]
    READY_FIELD_NUMBER: _ClassVar[int]
    ready: bool
    def __init__(self, ready: bool = ...) -> None: ...

class RepositoryModelLoadRequest(_message.Message):
    __slots__ = ["model_name"]
    MODEL_NAME_FIELD_NUMBER: _ClassVar[int]
    model_name: str
    def __init__(self, model_name: _Optional[str] = ...) -> None: ...

class RepositoryModelLoadResponse(_message.Message):
    __slots__ = ["isLoaded", "model_name"]
    ISLOADED_FIELD_NUMBER: _ClassVar[int]
    MODEL_NAME_FIELD_NUMBER: _ClassVar[int]
    isLoaded: bool
    model_name: str
    def __init__(self, model_name: _Optional[str] = ..., isLoaded: bool = ...) -> None: ...

class RepositoryModelUnloadRequest(_message.Message):
    __slots__ = ["model_name"]
    MODEL_NAME_FIELD_NUMBER: _ClassVar[int]
    model_name: str
    def __init__(self, model_name: _Optional[str] = ...) -> None: ...

class RepositoryModelUnloadResponse(_message.Message):
    __slots__ = ["isUnloaded", "model_name"]
    ISUNLOADED_FIELD_NUMBER: _ClassVar[int]
    MODEL_NAME_FIELD_NUMBER: _ClassVar[int]
    isUnloaded: bool
    model_name: str
    def __init__(self, model_name: _Optional[str] = ..., isUnloaded: bool = ...) -> None: ...

class ServerLiveRequest(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class ServerLiveResponse(_message.Message):
    __slots__ = ["live"]
    LIVE_FIELD_NUMBER: _ClassVar[int]
    live: bool
    def __init__(self, live: bool = ...) -> None: ...

class ServerMetadataRequest(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class ServerMetadataResponse(_message.Message):
    __slots__ = ["extensions", "name", "version"]
    EXTENSIONS_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    extensions: _containers.RepeatedScalarFieldContainer[str]
    name: str
    version: str
    def __init__(self, name: _Optional[str] = ..., version: _Optional[str] = ..., extensions: _Optional[_Iterable[str]] = ...) -> None: ...

class ServerReadyRequest(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class ServerReadyResponse(_message.Message):
    __slots__ = ["ready"]
    READY_FIELD_NUMBER: _ClassVar[int]
    ready: bool
    def __init__(self, ready: bool = ...) -> None: ...



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from kserve.logging import logger

try:
    from ray.serve.handle import DeploymentHandle, DeploymentResponse
except ImportError:
    logger.error(
        "Ray dependency is missing. Install Ray Serve with: pip install kserve[ray]"
    )
    raise
from typing import cast, Dict, Union, Optional
from cloudevents.http import CloudEvent

from .model import InferenceModel, InferenceVerb, InferReturnType
from .protocol.infer_type import InferRequest


class RayModel(InferenceModel):
    """
    Wrapper for a model that is deployed with Ray. All calls are delegated to the DeploymentHandle.
    """

    def __init__(self, name: str, handle: DeploymentHandle):
        super().__init__(name)
        self._handle = handle

    def __call__(
        self,
        body: Union[Dict, CloudEvent, InferRequest],
        headers: Optional[Dict[str, str]] = None,
        verb: InferenceVerb = InferenceVerb.PREDICT,
    ) -> InferReturnType:
        return cast(DeploymentResponse, self._handle.remote(body, headers=headers))

    def load(self):
        self._handle.load.remote()
        self.ready = True

    async def get_input_types(self):
        return await self._handle.get_input_types.remote()

    async def get_output_types(self):
        return await self._handle.get_output_types.remote()



# Copyright 2024 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import AsyncIterator, Optional, Union
import httpx
from http import HTTPStatus
from functools import partial, wraps
import orjson

from pydantic import ValidationError


from .openai_model import (
    BaseOpenAIRequest,
    OpenAICompletionModel,
    AsyncMappingIterator,
    CompletionRequest,
    ChatCompletionRequest,
)
from .types import (
    ChatCompletion,
    ChatCompletionChunk,
    Completion,
    ErrorResponse,
)
from .errors import OpenAIError, create_error_response
from ....logging import logger


COMPLETIONS_ENDPOINT = "/v1/completions"
CHAT_COMPLETIONS_ENDPOINT = "/v1/chat/completions"


def error_handler(f):
    @wraps(f)
    async def wrapper(*args, **kwargs):
        try:
            res = await f(*args, **kwargs)
            return res
        except httpx.HTTPStatusError as e:
            try:
                # Try to parse upstream error as an ErrorResponse object
                response = ErrorResponse.model_validate_json(e.response.content)
            except ValidationError:
                logger.warning(
                    f"Failed to parse error response from upstream: {e.response.content}"
                )
                response = create_error_response(
                    f"Received invalid response from upstream: {e.response.text}",
                    status_code=HTTPStatus.BAD_GATEWAY,
                    err_type="BadGateway",
                )
            raise OpenAIError(response=response)

        except httpx.TimeoutException as e:
            raise OpenAIError(
                response=create_error_response(
                    f"Timed out when communicating with upstream: {e}",
                    err_type="GatewayTimeout",
                    status_code=HTTPStatus.GATEWAY_TIMEOUT,
                )
            )
        except httpx.NetworkError as e:
            raise OpenAIError(
                response=create_error_response(
                    f"Failed to communicate with upstream: {e}",
                    err_type="ServiceUnavailableError",
                    status_code=HTTPStatus.SERVICE_UNAVAILABLE,
                )
            )
        except httpx.HTTPError as e:
            raise OpenAIError(
                response=create_error_response(
                    f"Upstream request failed: {e}",
                    err_type="InternalServerError",
                    status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
                )
            )

    return wrapper


class OpenAIProxyModel(OpenAICompletionModel):
    """
    An implementation of OpenAIModel that proxies requests to a backend server exposing Open AI endpoints.

    Users can extend this class and override any of the following methods to hook into the request/response cycle:
        - preprocess_completion_request
        - postprocess_completion
        - postprocess_completion_chunk
        - preprocess_chat_completion_request
        - postprocess_chat_completion
        - postprocess_chat_completion_chunk

    Each method takes a single parameter: the object currently being processed. Request objects may be mutated to modify
    the request sent to the downstream server and response objects (completions/completion chunks) may be mutated to
    modify the response returned to the upstream caller.

    Args:
        predictor_url (str):
            The url of the model server to send requests to. Should be fully qualified with scheme, host, and port.
            e.g. `http://my-backend:9000`
        http_client (httpx.AsyncClient|None):
            An optional instance of httpx.AsyncClient to use for sending requests to the upstream server.
    """

    predictor_url: str
    skip_upstream_validation: bool
    _http_client: httpx.AsyncClient
    _completions_endpoint: str
    _chat_completions_endpoint: str

    def __init__(
        self,
        name: str,
        predictor_url: str,
        http_client: Optional[httpx.AsyncClient] = None,
        skip_upstream_validation: bool = False,
    ):
        super().__init__(name)
        self.predictor_url = predictor_url
        self._http_client = (
            http_client
            if http_client is not None
            else httpx.AsyncClient(timeout=httpx.Timeout(timeout=5.0, read=30.0))
        )
        self._completions_endpoint = (
            f"{self.predictor_url.rstrip('/')}{COMPLETIONS_ENDPOINT}"
        )
        self._chat_completions_endpoint = (
            f"{self.predictor_url.rstrip('/')}{CHAT_COMPLETIONS_ENDPOINT}"
        )
        self.skip_upstream_validation = skip_upstream_validation
        self.ready = True

    def preprocess_completion_request(self, request: CompletionRequest):
        """Preprocess a completion request."""
        pass

    def postprocess_completion(
        self, completion: Completion, request: CompletionRequest
    ):
        """Postprocess a completion. Only called when response is not being streamed (i.e. stream=false)"""
        pass

    def postprocess_completion_chunk(
        self, completion: Completion, request: CompletionRequest
    ):
        """Postprocess a completion chunk. Only called when response is being streamed (i.e. stream=true)
        This method will be called once for each chunk that is streamed back to the user.
        """
        pass

    def preprocess_chat_completion_request(self, request: ChatCompletionRequest):
        """Preprocess a chat completion request."""
        pass

    def postprocess_chat_completion(
        self, chat_completion: ChatCompletion, request: ChatCompletionRequest
    ):
        """Postprocess a chat completion. Only called when response is not being streamed (i.e. stream=false)"""
        pass

    def postprocess_chat_completion_chunk(
        self, chat_completion_chunk: ChatCompletionChunk, request: ChatCompletionRequest
    ):
        """Postprocess a chat completion chunk. Only called when response is being streamed (i.e. stream=true)
        This method will be called once for each chunk that is streamed back to the user.
        """
        pass

    def _handle_completion_chunk(self, raw_chunk: str, request: CompletionRequest):
        # Skip empty lines
        if len(raw_chunk) == 0:
            return None
        # All chunks are prefixed with "data: "
        data = raw_chunk[6:]
        if data == "[DONE]":
            return None

        if self.skip_upstream_validation:
            obj = orjson.loads(data)
            completion_chunk = Completion.model_construct(**obj)
        else:
            completion_chunk = Completion.model_validate_json(data)
        self.postprocess_completion_chunk(completion_chunk, request)
        return completion_chunk

    def _handle_chat_completion_chunk(
        self, raw_chunk: str, request: ChatCompletionRequest
    ):
        # Skip empty lines
        if len(raw_chunk) == 0:
            return None
        # All chunks are prefixed with "data: "
        if len(raw_chunk) == 0:
            return None
        data = raw_chunk[6:]
        if data == "[DONE]":
            return None
        if self.skip_upstream_validation:
            obj = orjson.loads(data)
            chat_completion_chunk = ChatCompletionChunk.model_construct(**obj)
        else:
            chat_completion_chunk = ChatCompletionChunk.model_validate_json(data)
        self.postprocess_chat_completion_chunk(chat_completion_chunk, request)
        return chat_completion_chunk

    def _build_request(
        self, endpoint: str, request: BaseOpenAIRequest
    ) -> httpx.Request:

        if request.context and "upstream_headers" in request.context:
            headers = httpx.Headers(request.context["upstream_headers"])
        else:
            headers = httpx.Headers()

        headers["Content-type"] = "application/json"

        req = self._http_client.build_request(
            "POST",
            endpoint,
            content=request.params.model_dump_json(
                exclude_unset=True, exclude_none=True
            ),
            headers=headers,
        )
        return req

    @error_handler
    async def create_completion(
        self, request: CompletionRequest
    ) -> Union[Completion, AsyncIterator[Completion]]:
        self.preprocess_completion_request(request)
        if request.params.stream:
            req = self._build_request(self._completions_endpoint, request)
            r = await self._http_client.send(req, stream=True)
            r.raise_for_status()
            it = AsyncMappingIterator(
                iterator=r.aiter_lines(),
                mapper=partial(self._handle_completion_chunk, request=request),
                close=r.aclose,
            )
            return it
        else:
            completion = await self.generate_completion(request)
            self.postprocess_completion(completion, request)
            return completion

    async def generate_completion(self, request: CompletionRequest) -> Completion:
        req = self._build_request(self._completions_endpoint, request)
        response = await self._http_client.send(req)
        response.raise_for_status()
        if self.skip_upstream_validation:
            obj = response.json()
            completion = Completion.model_construct(**obj)
        else:
            completion = Completion.model_validate_json(response.content)
        return completion

    @error_handler
    async def create_chat_completion(
        self, request: ChatCompletionRequest
    ) -> Union[ChatCompletion, AsyncIterator[ChatCompletionChunk]]:
        self.preprocess_chat_completion_request(request)
        if request.params.stream:
            req = self._build_request(self._chat_completions_endpoint, request)
            r = await self._http_client.send(req, stream=True)
            r.raise_for_status()
            it = AsyncMappingIterator(
                iterator=r.aiter_lines(),
                mapper=partial(self._handle_chat_completion_chunk, request=request),
                close=r.aclose,
            )
            return it
        else:
            chat_completion = await self.generate_chat_completion(request)
            self.postprocess_chat_completion(chat_completion, request)
            return chat_completion

    async def generate_chat_completion(
        self, request: ChatCompletionRequest
    ) -> ChatCompletion:
        req = self._build_request(self._chat_completions_endpoint, request)
        response = await self._http_client.send(req)
        response.raise_for_status()
        if self.skip_upstream_validation:
            obj = response.json()
            chat_completion = ChatCompletion.model_construct(**obj)
        else:
            chat_completion = ChatCompletion.model_validate_json(response.content)
        return chat_completion



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import uuid

from kserve.protocol.grpc.grpc_predict_v2_pb2 import InferParameter
from typing import Dict, Union, List

from kserve.utils.numpy_codec import from_np_dtype
import pandas as pd
import numpy as np
import psutil
from cloudevents.conversion import to_binary, to_structured
from cloudevents.http import CloudEvent
from grpc import ServicerContext
from kserve.protocol.infer_type import InferOutput, InferRequest, InferResponse
from ..constants.constants import PredictorProtocol
from ..errors import InvalidInput


def is_running_in_k8s():
    return os.path.isdir("/var/run/secrets/kubernetes.io/")


def get_current_k8s_namespace():
    with open("/var/run/secrets/kubernetes.io/serviceaccount/namespace", "r") as f:
        return f.readline()


def get_default_target_namespace():
    if not is_running_in_k8s():
        return "default"
    return get_current_k8s_namespace()


def get_isvc_namespace(inferenceservice):
    return inferenceservice.metadata.namespace or get_default_target_namespace()


def get_ig_namespace(inferencegraph):
    return inferencegraph.metadata.namespace or get_default_target_namespace()


def cpu_count():
    """Get the available CPU count for this system.
    Takes the minimum value from the following locations:
    - Total system cpus available on the host.
    - CPU Affinity (if set)
    - Cgroups limit (if set)
    """
    count = os.cpu_count()

    # Check CPU affinity if available
    try:
        affinity_count = len(psutil.Process().cpu_affinity())
        if affinity_count > 0:
            count = min(count, affinity_count)
    except Exception:
        pass

    # Check cgroups if available
    if sys.platform == "linux":
        try:
            with open("/sys/fs/cgroup/cpu,cpuacct/cpu.cfs_quota_us") as f:
                quota = int(f.read())
            with open("/sys/fs/cgroup/cpu,cpuacct/cpu.cfs_period_us") as f:
                period = int(f.read())
            cgroups_count = int(quota / period)
            if cgroups_count > 0:
                count = min(count, cgroups_count)
        except Exception:
            pass

    return count


def is_structured_cloudevent(body: Dict) -> bool:
    """Returns True if the JSON request body resembles a structured CloudEvent"""
    return (
        "time" in body
        and "type" in body
        and "source" in body
        and "id" in body
        and "specversion" in body
        and "data" in body
    )


def create_response_cloudevent(
    model_name: str, response: Dict, req_attributes: Dict, binary_event=False
) -> tuple:
    ce_attributes = {}

    if os.getenv("CE_MERGE", "false").lower() == "true":
        if binary_event:
            ce_attributes = req_attributes
            if "datacontenttype" in ce_attributes:  # Optional field so must check
                del ce_attributes["datacontenttype"]
        else:
            ce_attributes = req_attributes

        # Remove these fields so we generate new ones
        del ce_attributes["id"]
        del ce_attributes["time"]

    ce_attributes["type"] = os.getenv("CE_TYPE", "io.kserve.inference.response")
    ce_attributes["source"] = os.getenv(
        "CE_SOURCE", f"io.kserve.inference.{model_name}"
    )

    event = CloudEvent(ce_attributes, response)

    if binary_event:
        event_headers, event_body = to_binary(event)
    else:
        event_headers, event_body = to_structured(event)

    return event_headers, event_body


def generate_uuid() -> str:
    return str(uuid.uuid4())


def to_headers(context: ServicerContext) -> Dict[str, str]:
    metadata = context.invocation_metadata()
    if hasattr(context, "trailing_metadata"):
        metadata += context.trailing_metadata()
    headers = {}
    for metadatum in metadata:
        headers[metadatum.key] = metadatum.value

    return headers


def get_predict_input(
    payload: Union[Dict, InferRequest], columns: List = None
) -> Union[np.ndarray, pd.DataFrame, List[str]]:
    if isinstance(payload, Dict):
        instances = payload["inputs"] if "inputs" in payload else payload["instances"]
        if len(instances) == 0:
            return np.array(instances)
        if isinstance(instances[0], Dict) or (
            isinstance(instances[0], List)
            and len(instances[0]) != 0
            and isinstance(instances[0][0], Dict)
        ):
            dfs = []
            for instance in instances:
                dfs.append(pd.DataFrame(instance, columns=columns))
            inputs = pd.concat(dfs, axis=0)
            return inputs
        else:
            if isinstance(instances[0], str):
                return instances
            return np.array(instances)
    elif isinstance(payload, InferRequest):
        content_type = ""
        parameters = payload.parameters
        if parameters:
            if isinstance(parameters.get("content_type"), InferParameter):
                # for v2 grpc, we get InferParameter obj eg: {"content_type": string_param: "pd"}
                content_type = str(parameters.get("content_type").string_param)
            else:
                # for v2 http, we get string eg: {"content_type": "pd"}
                content_type = parameters.get("content_type")

        if content_type == "pd":
            return payload.as_dataframe()
        else:
            input = payload.inputs[0]
            if (
                input.datatype == "BYTES"
                and len(input.data) > 0
                and isinstance(input.data[0], str)
            ):
                return input.data
            return input.as_numpy()


def get_predict_response(
    payload: Union[Dict, InferRequest],
    result: Union[np.ndarray, List, pd.DataFrame],
    model_name: str,
) -> Union[Dict, InferResponse]:
    if isinstance(payload, Dict):
        infer_outputs = result
        if isinstance(result, pd.DataFrame):
            infer_outputs = []
            for label, row in result.iterrows():
                infer_outputs.append(row.to_dict())
        elif isinstance(result, np.ndarray):
            infer_outputs = result.tolist()
        return {"predictions": infer_outputs}
    elif isinstance(payload, InferRequest):
        infer_outputs = []
        if isinstance(result, pd.DataFrame):
            for col in result.columns:
                infer_output = InferOutput(
                    name=col,
                    shape=list(result[col].shape),
                    datatype=from_np_dtype(result[col].dtype),
                )
                infer_output.set_data_from_numpy(
                    result[col].to_numpy(), binary_data=payload.use_binary_outputs
                )
                infer_outputs.append(infer_output)
        elif (
            isinstance(result, list) and len(result) > 0 and isinstance(result[0], str)
        ):
            infer_output = InferOutput(
                name="output-0",
                shape=[len(result)],
                datatype="BYTES",
            )
            infer_output.set_data_from_numpy(
                np.array(result, dtype=np.object_),
                binary_data=payload.use_binary_outputs,
            )
            infer_outputs.append(infer_output)
        else:
            if isinstance(result, list):
                result = np.array(result)
            infer_output = InferOutput(
                name="output-0",
                shape=list(result.shape),
                datatype=from_np_dtype(result.dtype),
            )
            infer_output.set_data_from_numpy(
                result, binary_data=payload.use_binary_outputs
            )
            infer_outputs.append(infer_output)
        return InferResponse(
            model_name=model_name,
            infer_outputs=infer_outputs,
            response_id=payload.id if payload.id else generate_uuid(),
            use_binary_outputs=payload.use_binary_outputs,
            requested_outputs=payload.request_outputs,
        )
    else:
        raise InvalidInput(f"unsupported payload type {type(payload)}")


def strtobool(val: str) -> bool:
    """Convert a string representation of truth to True or False.

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.

    Adapted from deprecated `distutils`
    https://github.com/python/cpython/blob/3.11/Lib/distutils/util.py
    """
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return True
    elif val in ("n", "no", "f", "false", "off", "0"):
        return False
    else:
        raise ValueError("invalid truth value %r" % (val,))


def is_v2(protocol: Union[str, PredictorProtocol]) -> bool:
    return protocol == PredictorProtocol.REST_V2 or (
        isinstance(protocol, str)
        and protocol.lower() == PredictorProtocol.REST_V2.value.lower()
    )


def is_v1(protocol: Union[str, PredictorProtocol]) -> bool:
    return protocol == PredictorProtocol.REST_V1 or (
        isinstance(protocol, str)
        and protocol.lower() == PredictorProtocol.REST_V1.value.lower()
    )



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import multiprocessing
from concurrent import futures

from grpc import aio

from kserve.logging import logger
from kserve.protocol.dataplane import DataPlane
from kserve.protocol.model_repository_extension import ModelRepositoryExtension

from . import grpc_predict_v2_pb2_grpc
from .interceptors import LoggingInterceptor, ExceptionToStatusInterceptor
from .servicer import InferenceServicer


class GRPCServer:
    def __init__(
        self,
        port: int,
        data_plane: DataPlane,
        model_repository_extension: ModelRepositoryExtension,
        kwargs: dict,
    ):
        self._port = port
        self._data_plane = data_plane
        self._model_repository_extension = model_repository_extension
        self._server = None
        self._kwargs = kwargs

    async def start(self, max_workers):
        inference_servicer = InferenceServicer(
            self._data_plane, self._model_repository_extension
        )
        self._server = aio.server(
            futures.ThreadPoolExecutor(max_workers=max_workers),
            interceptors=(LoggingInterceptor(), ExceptionToStatusInterceptor()),
            options=[
                (
                    "grpc.max_send_message_length",
                    self._kwargs.get("grpc_max_send_message_length"),
                ),
                (
                    "grpc.max_receive_message_length",
                    self._kwargs.get("grpc_max_receive_message_length"),
                ),
            ],
        )
        grpc_predict_v2_pb2_grpc.add_GRPCInferenceServiceServicer_to_server(
            inference_servicer, self._server
        )

        listen_addr = f"[::]:{self._port}"
        self._server.add_insecure_port(listen_addr)
        logger.info(f"Starting gRPC server with {max_workers} workers")
        logger.info("Starting gRPC server on %s", listen_addr)
        await self._server.start()
        await self._server.wait_for_termination()

    async def stop(self, sig: int = None):
        if self._server:
            logger.info("Waiting for gRPC server shutdown")
            await self._server.stop(grace=10)
            logger.info("gRPC server shutdown complete")


class GRPCProcess(multiprocessing.Process):

    def __init__(
        self,
        port: int,
        max_threads: int,
        data_plane: DataPlane,
        model_repository_extension: ModelRepositoryExtension,
    ):
        super().__init__()
        self._data_plane = data_plane
        self._model_repository_extension = model_repository_extension
        self._port = port
        self._max_threads = max_threads
        self._server = None

    def stop(self):
        self._server.stop()

    def run(self):
        self._server = GRPCServer(
            self._port, self._data_plane, self._model_repository_extension
        )
        asyncio.run(self._server.start(self._max_threads))



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import sys
from typing import Dict, List, Optional

from ..errors import ModelNotFound, ModelNotReady
from ..model_repository import ModelRepository


class ModelRepositoryExtension:
    """
    This is a class implements the 'Model Repository Extension' to the kserve Protocol V2 as
    described in 'https://github.com/triton-inference-server/server/blob/main/docs/protocol
    /extension_model_repository.md#model-repository-extension'

    Attributes:
        model_registry (ModelRepository): Backing model store
    """

    def __init__(self, model_registry: ModelRepository):
        self._model_registry = model_registry

    async def index(self, filter_ready: Optional[bool] = False) -> List[Dict[str, str]]:
        """Returns information about every model available in a model repository.

        Args:
            filter_ready: When set True, the function returns only the models that are ready

        Returns:
            List[Dict[str, str]]: list with metadata for models as below:

                {
                    name: model_name,
                    state: "Ready" or "NotReady"
                    reason: ""
                }
        """
        model_list = []
        for model_name in self._model_registry.get_models().keys():
            model_ready = await self._model_registry.is_model_ready(model_name)
            if model_ready or not filter_ready:
                # If model is ready or filter_ready is set to False
                model_list.append(
                    {
                        "name": model_name,
                        "state": ("Ready" if model_ready else "NotReady"),
                        "reason": "",
                    }
                )

        return model_list

    async def load(self, model_name: str) -> None:
        """Loads the specified model.

        Args:
            model_name (str): name of the model to load.

        Returns: None

        Raises:
            ModelNotReady: Exception if model loading fails.
        """
        try:
            # For backward compatibility, the synchronous `load` has been kept here.
            if inspect.iscoroutinefunction(self._model_registry.load):
                await self._model_registry.load(model_name)
            else:
                self._model_registry.load(model_name)
        except Exception:
            ex_type, ex_value, ex_traceback = sys.exc_info()
            raise ModelNotReady(
                model_name, f"Error type: {ex_type} error msg: {ex_value}"
            )
        is_ready = await self._model_registry.is_model_ready(model_name)
        if not is_ready:
            raise ModelNotReady(model_name)

    async def unload(self, model_name: str) -> None:
        """Unload the specified model.

        Args:
            model_name (str): Name of the model to unload.

        Returns: None

        Raises:
            ModelNotFound: Exception if the requested model is not found.
        """
        try:
            self._model_registry.unload(model_name)
        except KeyError:
            raise ModelNotFound(model_name)



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
    KServe

    Python SDK for KServe  # noqa: E501

    The version of the OpenAPI document: v0.1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from kserve.configuration import Configuration


class V1beta1ModelStatus(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'copies': 'V1beta1ModelCopies',
        'last_failure_info': 'V1beta1FailureInfo',
        'states': 'V1beta1ModelRevisionStates',
        'transition_status': 'str'
    }

    attribute_map = {
        'copies': 'copies',
        'last_failure_info': 'lastFailureInfo',
        'states': 'states',
        'transition_status': 'transitionStatus'
    }

    def __init__(self, copies=None, last_failure_info=None, states=None, transition_status='', local_vars_configuration=None):  # noqa: E501
        """V1beta1ModelStatus - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._copies = None
        self._last_failure_info = None
        self._states = None
        self._transition_status = None
        self.discriminator = None

        if copies is not None:
            self.copies = copies
        if last_failure_info is not None:
            self.last_failure_info = last_failure_info
        if states is not None:
            self.states = states
        self.transition_status = transition_status

    @property
    def copies(self):
        """Gets the copies of this V1beta1ModelStatus.  # noqa: E501


        :return: The copies of this V1beta1ModelStatus.  # noqa: E501
        :rtype: V1beta1ModelCopies
        """
        return self._copies

    @copies.setter
    def copies(self, copies):
        """Sets the copies of this V1beta1ModelStatus.


        :param copies: The copies of this V1beta1ModelStatus.  # noqa: E501
        :type: V1beta1ModelCopies
        """

        self._copies = copies

    @property
    def last_failure_info(self):
        """Gets the last_failure_info of this V1beta1ModelStatus.  # noqa: E501


        :return: The last_failure_info of this V1beta1ModelStatus.  # noqa: E501
        :rtype: V1beta1FailureInfo
        """
        return self._last_failure_info

    @last_failure_info.setter
    def last_failure_info(self, last_failure_info):
        """Sets the last_failure_info of this V1beta1ModelStatus.


        :param last_failure_info: The last_failure_info of this V1beta1ModelStatus.  # noqa: E501
        :type: V1beta1FailureInfo
        """

        self._last_failure_info = last_failure_info

    @property
    def states(self):
        """Gets the states of this V1beta1ModelStatus.  # noqa: E501


        :return: The states of this V1beta1ModelStatus.  # noqa: E501
        :rtype: V1beta1ModelRevisionStates
        """
        return self._states

    @states.setter
    def states(self, states):
        """Sets the states of this V1beta1ModelStatus.


        :param states: The states of this V1beta1ModelStatus.  # noqa: E501
        :type: V1beta1ModelRevisionStates
        """

        self._states = states

    @property
    def transition_status(self):
        """Gets the transition_status of this V1beta1ModelStatus.  # noqa: E501

        Whether the available predictor endpoints reflect the current Spec or is in transition  # noqa: E501

        :return: The transition_status of this V1beta1ModelStatus.  # noqa: E501
        :rtype: str
        """
        return self._transition_status

    @transition_status.setter
    def transition_status(self, transition_status):
        """Sets the transition_status of this V1beta1ModelStatus.

        Whether the available predictor endpoints reflect the current Spec or is in transition  # noqa: E501

        :param transition_status: The transition_status of this V1beta1ModelStatus.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and transition_status is None:  # noqa: E501
            raise ValueError("Invalid value for `transition_status`, must not be `None`")  # noqa: E501

        self._transition_status = transition_status

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1beta1ModelStatus):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1beta1ModelStatus):
            return True

        return self.to_dict() != other.to_dict()
